"""Module providing classes to establish ssh or ssh-like connections to hosts.

Classes:
    Session - Wrapper around ssh2-python's auth/connection system.
    InteractiveShell - Wrapper around ssh2-python's non-blocking channel system.

Note: You typically want to use a Host object instance to create sessions,
      not these classes directly.
"""

from contextlib import contextmanager
from pathlib import Path

from logzero import logger
from ssh2 import sftp as _sftp
from ssh2.exceptions import SocketSendError
from ssh2.session import Session as _Session

from broker import exceptions, helpers
from broker.binds.utils import _create_connect_socket

SFTP_MODE = (
    _sftp.LIBSSH2_SFTP_S_IRUSR
    | _sftp.LIBSSH2_SFTP_S_IWUSR
    | _sftp.LIBSSH2_SFTP_S_IRGRP
    | _sftp.LIBSSH2_SFTP_S_IROTH
)
FILE_FLAGS = _sftp.LIBSSH2_FXF_CREAT | _sftp.LIBSSH2_FXF_WRITE | _sftp.LIBSSH2_FXF_TRUNC


class Session:
    """Wrapper around ssh2-python's auth/connection system."""

    def __init__(self, **kwargs):
        """Initialize a Session object.

        kwargs:
            hostname (str): The hostname or IP address of the remote host. Defaults to 'localhost'.
            username (str): The username to authenticate with. Defaults to 'root'.
            timeout (float): The timeout for the connection in seconds. Defaults to 60.
            port (int): The port number to connect to. Defaults to 22.
            key_filename (str): The path to the private key file to use for authentication.
            password (str): The password to use for authentication.
            ipv6 (bool): Whether or not to use IPv6. Defaults to False.
            ipv4_fallback (bool): Whether or not to fallback to IPv4 if IPv6 fails. Defaults to True.

        Raises:
            AuthException: If no password or key file is provided.
            ConnectionError: If the connection fails.
            FileNotFoundError: If the key file is not found.
        """
        host = kwargs.get("hostname", "localhost")
        user = kwargs.get("username", "root")
        port = kwargs.get("port", 22)
        key_filename = kwargs.get("key_filename")
        password = kwargs.get("password")
        timeout = kwargs.get("timeout", 60)

        # Create the socket
        self.sock, self.is_ipv6 = _create_connect_socket(
            host,
            port,
            timeout,
            ipv6=kwargs.get("ipv6", False),
            ipv4_fallback=kwargs.get("ipv4_fallback", True),
        )

        self.session = _Session()

        self.session.handshake(self.sock)
        try:
            if key_filename:
                auth_type = "Key"
                if not Path(key_filename).exists():
                    raise FileNotFoundError(f"Key not found in '{key_filename}'")
                self.session.userauth_publickey_fromfile(user, key_filename)
            elif password:
                auth_type = "Password"
                self.session.userauth_password(user, password)
            elif user:
                auth_type = "Session"
                self.session.agent_auth(user)
            else:
                raise exceptions.AuthenticationError("No password or key file provided.")
        except Exception as err:
            raise exceptions.AuthenticationError(
                f"{auth_type}-based authentication failed."
            ) from err

    @staticmethod
    def _read(channel):
        """Read the contents of a channel."""
        size, data = channel.read()
        results = ""
        while size > 0:
            try:
                results += data.decode("utf-8")
            except UnicodeDecodeError as err:
                logger.error(f"Skipping data chunk due to {err}\nReceived: {data}")
            size, data = channel.read()
        return helpers.Result.from_ssh(
            stdout=results,
            channel=channel,
        )

    @staticmethod
    def _set_destination(source, destination):
        dest = destination or source
        if dest.endswith("/"):
            dest = dest + Path(source).name
        return dest

    def disconnect(self):
        """Disconnect session."""
        self.session.disconnect()

    def remote_copy(self, source, dest_host, dest_path=None, ensure_dir=True):
        """Copy a file from this host to another."""
        dest_path = dest_path or source
        sftp_down = self.session.sftp_init()
        sftp_up = dest_host.session.session.sftp_init()
        if ensure_dir:
            dest_host.session.run(f"mkdir -p {Path(dest_path).absolute().parent}")
        with (
            sftp_down.open(source, _sftp.LIBSSH2_FXF_READ, _sftp.LIBSSH2_SFTP_S_IRUSR) as download,
            sftp_up.open(dest_path, FILE_FLAGS, SFTP_MODE) as upload,
        ):
            for _size, data in download:
                upload.write(data)

    def run(self, command, timeout=0):
        """Run a command on the host and return the results."""
        self.session.set_timeout(helpers.translate_timeout(timeout))
        try:
            channel = self.session.open_session()
        except SocketSendError as err:
            logger.warning(
                f"Encountered connection issue. Attempting to reconnect and retry.\n{err}"
            )
            # FIXME _session is on the Host, not Session
            del self._session
            channel = self.session.open_session()
        channel.execute(
            command,
        )
        channel.wait_eof()
        channel.close()
        channel.wait_closed()
        results = self._read(channel)
        return results

    def scp_write(self, source, destination=None, ensure_dir=True):
        """SCP write a local file to a remote destination."""
        destination = self._set_destination(source, destination)
        source_path = Path(source)  # Store the Path object
        fileinfo = source_path.stat()

        chan = self.session.scp_send64(
            destination,
            fileinfo.st_mode & 0o777,
            fileinfo.st_size,
            fileinfo.st_mtime,
            fileinfo.st_atime,
        )
        if ensure_dir:
            self.run(f"mkdir -p {Path(destination).absolute().parent}")
        with source_path.open("rb") as local:
            for data in local:
                chan.write(data)

    def sftp_read(self, source, destination=None, return_data=False):
        """Read a remote file into a local destination or return a bytes object if return_data is True."""
        if not return_data:
            destination = self._set_destination(source, destination)

            # create the destination path if it doesn't exist
            destination = Path(destination)
            destination.parent.mkdir(parents=True, exist_ok=True)

        # initiate the sftp session, read data, write it to a local destination
        sftp = self.session.sftp_init()
        with sftp.open(source, _sftp.LIBSSH2_FXF_READ, _sftp.LIBSSH2_SFTP_S_IRUSR) as remote:
            captured_data = b""
            for _rc, data in remote:
                captured_data += data
            if return_data:
                return captured_data
            destination.write_bytes(captured_data)

    def sftp_write(self, source, destination=None, ensure_dir=True):
        """Sftp write a local file to a remote destination."""
        destination = self._set_destination(source, destination)

        data = Path(source).read_bytes()
        if ensure_dir:
            self.run(f"mkdir -p {Path(destination).absolute().parent}")

        sftp = self.session.sftp_init()
        with sftp.open(destination, FILE_FLAGS, SFTP_MODE) as remote:
            remote.write(data)

    def shell(self, pty=False):
        """Create and return an interactive shell instance."""
        channel = self.session.open_session()
        return InteractiveShell(channel, pty)

    @contextmanager
    def tail_file(self, filename):
        """Simulate tailing a file on the remote host.

        Example:
            with my_host.session.tail_file("/var/log/messages") as res:
                # do something that creates new messages
            print(res.stdout)

        yields a FileTailer object with contents attr set to the string output
        """
        # TODO refactor to use SFTP instead
        initial_size = int(self.run(f"stat -c %s {filename}").stdout.strip())
        yield (tailer := FileTailer(session=self.session, filename=filename))
        tailer.contents = self.run(f"tail -c +{initial_size} {filename}").stdout


class FileTailer:
    """FileTailer class."""

    def __init__(self, **kwargs):
        self.session = kwargs.get("session")
        self.filename = kwargs.get("filename")


class InteractiveShell:
    """A helper class that provides an interactive shell interface.

    Preferred use of this class is via its context manager

    with InteractiveShell(channel=my_channel) as shell:
        shell.send("some-command --argument")
        shell.send("another-command")
        time.sleep(5)  # give time for things to complete
    assert "expected text" in shell.result.stdout

    """

    def __init__(self, channel, pty=False):
        self._channel = channel
        if pty:
            self._channel.pty()
        self._channel.shell()

    def __enter__(self):
        """Return the shell object."""
        return self

    def __exit__(self, *exc_args):
        """Close the channel and read stdout/stderr and status."""
        self._channel.close()
        self.result = Session._read(self._channel)

    def __getattribute__(self, name):
        """Expose non-duplicate attributes from the channel."""
        try:
            return object.__getattribute__(self, name)
        except AttributeError:
            return getattr(self._channel, name)

    def send(self, cmd):
        """Send a command to the channel, ensuring a newline character."""
        if not cmd.endswith("\n"):
            cmd += "\n"
        self._channel.write(cmd)

    def stdout(self):
        """Read the contents of a channel's stdout."""
        if not self._channel.eof():
            _, data = self._channel.read(65535)
            results = data.decode("utf-8")
        else:
            results = None
            size, data = self._channel.read()
            while size > 0:
                results += data.decode("utf-8")
                size, data = self._channel.read()
        return results



"""A collection of classes to ease interaction with Docker and Podman libraries."""

from broker.exceptions import UserError

HEADER_SIZE = 8
STDOUT = 1
STDERR = 2
SSH_PORT = 22


def demux_output(data_bytes):
    """Demuxes the output of a container stream into stdout and stderr streams.

    Stream data is expected to be in the following format:
    - 1 byte: stream type (1=stdout, 2=stderr)
    - 3 bytes: padding
    - 4 bytes: payload size (big-endian)
    - N bytes: payload data
    ref: https://docs.podman.io/en/latest/_static/api.html?version=v5.0#tag/containers/operation/ContainerAttachLibpod

    Args:
        data_bytes: Bytes object containing the combined stream data.

    Returns:
        A tuple containing two bytes objects: (stdout, stderr).
    """
    stdout = b""
    stderr = b""
    while len(data_bytes) >= HEADER_SIZE:
        # Extract header information
        header, data_bytes = data_bytes[:HEADER_SIZE], data_bytes[HEADER_SIZE:]
        stream_type = header[0]
        payload_size = int.from_bytes(header[4:HEADER_SIZE], "big")
        # Check if data is sufficient for payload
        if len(data_bytes) < payload_size:
            break  # Incomplete frame, wait for more data

        # Extract and process payload
        payload = data_bytes[:payload_size]
        if stream_type == STDOUT:
            stdout += payload
        elif stream_type == STDERR:
            stderr += payload
        else:
            # todo: Handle unexpected stream types
            pass

        # Update data for next frame
        data_bytes = data_bytes[payload_size:]

    return stdout, stderr


class ContainerBind:
    """A base class that provides common functionality for Docker and Podman containers."""

    _sensitive_attrs = ["password", "host_password"]

    def __init__(self, host=None, username=None, password=None, port=22, timeout=None):
        self.host = host
        self.username = username
        self.password = password
        self.port = port
        self.timeout = timeout
        self._client = None
        self._ClientClass = None

    @property
    def client(self):
        """Return the client instance. Create one if it does not exist."""
        if not isinstance(self._client, self._ClientClass):
            self._client = self._ClientClass(base_url=self.uri, timeout=self.timeout)
        return self._client

    @property
    def images(self):
        """Return a list of images on the container host."""
        return self.client.images.list()

    @property
    def containers(self):
        """Return a list of containers on the container host."""
        return self.client.containers.list(all=True)

    @property
    def networks(self):
        """Return a list of networks on the container host."""
        return self.client.networks.list()

    def image_info(self, name):
        """Return curated information about an image on the container host."""
        if image := self.client.images.get(name):
            return {
                "id": image.short_id,
                "tags": image.tags,
                "size": image.attrs["Size"],
                "config": {k: v for k, v in image.attrs["Config"].items() if k != "Env"},
            }

    def create_container(self, image, command=None, net_name=None, **kwargs):
        """Create and return running container instance."""
        if net_name:
            net_dict = {}
            for name in net_name.split(","):
                if not self.get_network_by_attrs({"name": name}):
                    raise UserError(f"Network '{name}' not found on container host.")
                net_dict[name] = {"NetworkId": name}
            kwargs["networks"] = net_dict
        kwargs = self._sanitize_create_args(kwargs)
        return self.client.containers.create(image, command, **kwargs)

    def execute(self, image, command=None, remove=True, **kwargs):
        """Run a container and return the raw result."""
        return self.client.containers.run(image, command=command, remove=remove, **kwargs).decode()

    def remove_container(self, container=None):
        """Remove a container from the container host."""
        if container:
            container.remove(v=True, force=True)

    def pull_image(self, name):
        """Pull an image into the container host."""
        return self.client.images.pull(name)

    def get_network_by_attrs(self, attr_dict):
        """Return the first matching network that matches all attr_dict keys and values."""
        for network in self.networks:
            if all(network.attrs.get(k) == v for k, v in attr_dict.items()):
                return network

    @staticmethod
    def get_logs(container):
        """Return the logs from a container."""
        return "\n".join(x.decode() for x in container.logs(stream=False))

    @staticmethod
    def get_attrs(cont):
        """Return curated information about a container."""
        return {
            "id": cont.id,
            "image": cont.attrs.get("ImageName", cont.attrs["Image"]),
            "name": cont.name or cont.attrs["Names"][0],
            "container_config": cont.attrs.get("Config", {}),
            "host_config": cont.attrs.get("HostConfig", {}),
            "ports": cont.ports or cont.attrs.get("Ports"),
        }

    def __repr__(self):
        """Return a string representation of the object."""
        inner = ", ".join(
            f"{k}={'******' if k in self._sensitive_attrs and v else v}"
            for k, v in self.__dict__.items()
            if not k.startswith("_") and not callable(v)
        )
        return f"{self.__class__.__name__}({inner})"


class PodmanBind(ContainerBind):
    """Handles Podman-specific connection and implementation differences."""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        from podman import PodmanClient

        self._ClientClass = PodmanClient
        if self.host == "localhost":
            self.uri = "unix:///run/user/1000/podman/podman.sock"
        elif kwargs.get("port") == SSH_PORT:
            self.uri = "http+ssh://{username}@{host}:{port}/run/podman/podman.sock".format(**kwargs)
        else:
            self.uri = "tcp://{host}:{port}".format(**kwargs)

    def _sanitize_create_args(self, kwargs):
        from podman.domain.containers_create import CreateMixin

        # Temporarily remove mounts so CreateMixin won't strip them out
        mounts = kwargs.pop("mounts", None)

        try:
            CreateMixin._render_payload(kwargs)
        except TypeError as err:
            sanitized = (
                err.args[0]
                .replace("Unknown keyword argument(s): ", "")
                .replace("'", "")
                .split(" ,")
            )
            kwargs = {k: v for k, v in kwargs.items() if k not in sanitized}
            return self._sanitize_create_args(kwargs)
        # Restore mounts so Podman sees bind-mount list intact
        if mounts is not None:
            kwargs["mounts"] = mounts

        return kwargs


class DockerBind(ContainerBind):
    """Handles Docker-specific connection and implementation differences."""

    def __init__(self, port=2375, **kwargs):
        kwargs["port"] = port
        super().__init__(**kwargs)
        from docker import DockerClient

        self._ClientClass = DockerClient
        if self.host == "localhost":
            self.uri = "unix://var/run/docker.sock"
        elif kwargs.get("port") == SSH_PORT:
            self.uri = "ssh://{username}@{host}:{port}".format(**kwargs)
        else:
            self.uri = "tcp://{username}@{host}".format(**kwargs)

    def _sanitize_create_args(self, kwargs):
        from docker.models.containers import RUN_CREATE_KWARGS, RUN_HOST_CONFIG_KWARGS

        special_kwargs = ["ports", "volumes", "network", "networking_config", "mounts"]
        accepted_kwargs = RUN_HOST_CONFIG_KWARGS + RUN_CREATE_KWARGS + special_kwargs
        return {k: v for k, v in kwargs.items() if k in accepted_kwargs}



"""Module providing classes to establish ssh or ssh-like connections to hosts.

Classes:
    Session - Wrapper around Paramiko's auth/connection system.

Note: You typically want to use a Host object instance to create sessions,
      not these classes directly.
"""

from contextlib import contextmanager
from pathlib import Path
import shlex
import socket

import paramiko

from broker import exceptions, helpers


class Session:
    """Wrapper around Paramiko's auth/connection system."""

    def _load_private_key(self, key_filename, password):
        """Attempt to load a private key file, trying various types."""
        key_path_str = str(Path(key_filename).expanduser().resolve())
        key_types = [
            paramiko.RSAKey,
            paramiko.DSSKey,
            paramiko.ECDSAKey,
            paramiko.Ed25519Key,
        ]
        last_exception = None

        for key_type in key_types:
            try:
                return key_type.from_private_key_file(key_path_str, password=password)
            except paramiko.PasswordRequiredException:  # noqa: PERF203
                # If password is required but not provided (or incorrect), raise immediately
                raise exceptions.AuthenticationError(
                    f"Password required or incorrect for key file: {key_path_str}"
                )
            except paramiko.SSHException as e:
                # Store the exception and try the next key type
                last_exception = e
                continue  # Try next key type

        # If all key types failed
        raise exceptions.AuthenticationError(
            f"Unsupported or invalid key file: {key_path_str} - {last_exception}"
        ) from last_exception

    def _authenticate_with_agent(self):
        """Attempt authentication using the SSH agent."""
        try:
            agent = paramiko.Agent()
            agent_keys = agent.get_keys()
            if not agent_keys:
                raise exceptions.AuthenticationError("Agent available but has no keys.")
            for key in agent_keys:
                try:
                    self.transport.auth_publickey(self.username, key)
                    if self.transport.is_authenticated():
                        return  # Success
                except paramiko.AuthenticationException:  # noqa: PERF203
                    continue  # Try next agent key
            # If loop finishes without authenticating
            raise exceptions.AuthenticationError("Agent authentication failed (tried all keys).")
        except Exception as e:  # Catch agent connection errors, etc.
            raise exceptions.AuthenticationError(f"Agent authentication failed: {e}") from e

    def __init__(self, **kwargs):  # noqa: PLR0912, PLR0915 - TODO: refactor?
        """Initialize a Session object.

        kwargs:
            hostname (str): The hostname or IP address of the remote host. Defaults to 'localhost'.
            username (str): The username to authenticate with. Defaults to 'root'.
            timeout (float): The timeout for the connection in seconds. Defaults to 60.
            port (int): The port number to connect to. Defaults to 22.
            key_filename (str): The path to the private key file to use for authentication.
            password (str): The password to use for authentication.
        """
        self.hostname = kwargs.get("hostname", "localhost")
        self.username = kwargs.get("username", "root")
        self.port = kwargs.get("port", 22)
        self.timeout = kwargs.get("timeout", 60)
        self.key_filename = kwargs.get("key_filename")
        self.password = kwargs.get("password")

        try:
            # Establish socket connection
            sock = socket.create_connection((self.hostname, self.port), self.timeout)
        except OSError as e:
            raise exceptions.ConnectionError(f"Failed to connect socket: {e}") from e

        self.transport = paramiko.Transport(sock)
        try:
            # Start SSH transport, potentially verifying host key
            self.transport.start_client(timeout=self.timeout)
            self.transport.set_keepalive(30)  # Set keepalive interval

        except paramiko.SSHException as e:
            self.transport.close()
            raise exceptions.ConnectionError(f"Failed to start SSH transport: {e}") from e

        # Load host keys and implement AutoAddPolicy behavior
        try:
            host_keys_path = Path("~/.ssh/known_hosts").expanduser().resolve()
            host_keys = paramiko.util.load_host_keys(str(host_keys_path))
            server_key = self.transport.get_remote_server_key()
            key_entry = host_keys.lookup(self.hostname)

            if not key_entry or key_entry.get(server_key.get_name()) != server_key:
                # Auto-accept unknown host keys (like AutoAddPolicy)
                # This is appropriate for test environments and development
                host_keys.add(self.hostname, server_key.get_name(), server_key)
                try:
                    # Try to save the host key to known_hosts
                    host_keys_path.parent.mkdir(parents=True, exist_ok=True)
                    host_keys.save(str(host_keys_path))
                except (OSError, PermissionError):
                    # If we can't save, that's okay - proceed anyway
                    pass

        except FileNotFoundError:
            # known_hosts file doesn't exist, which is fine for first connection
            # Auto-accept the host key
            try:
                host_keys_path.parent.mkdir(parents=True, exist_ok=True)
                server_key = self.transport.get_remote_server_key()
                host_keys = paramiko.HostKeys()
                host_keys.add(self.hostname, server_key.get_name(), server_key)
                host_keys.save(str(host_keys_path))
            except (OSError, PermissionError):
                # If we can't save, that's okay - proceed anyway
                pass
        except (OSError, PermissionError, paramiko.SSHException):
            # For any other host key related errors, just proceed
            # This ensures compatibility with test environments
            pass

        # Attempt authentication
        auth_method = "none"
        try:
            if self.key_filename:
                auth_method = "key"
                key = self._load_private_key(self.key_filename, self.password)
                self.transport.auth_publickey(self.username, key)

            elif self.password:
                auth_method = "password"
                self.transport.auth_password(self.username, self.password)
            else:
                # Try agent authentication if no key/password provided
                auth_method = "agent"
                self._authenticate_with_agent()

            if not self.transport.is_authenticated():
                # This path should ideally not be reached if helpers raise correctly.
                raise exceptions.AuthenticationError(
                    f"{auth_method.capitalize()} authentication failed."
                )

            # Create an SSHClient interface attached to the authenticated transport
            self.client = paramiko.SSHClient()
            self.client._transport = self.transport  # Attach transport

        except (paramiko.AuthenticationException, exceptions.AuthenticationError) as err:
            self.transport.close()
            # Re-raise specific AuthenticationError or a general one if needed
            if isinstance(err, exceptions.AuthenticationError):
                raise  # Propagate our specific error
            else:
                raise exceptions.AuthenticationError(
                    f"{auth_method.capitalize()} authentication failed: {err}"
                ) from err
        except (
            FileNotFoundError
        ) as err:  # Should be caught by _load_private_key now, but keep for safety
            self.transport.close()
            raise exceptions.AuthenticationError(
                f"Key file not found: {self.key_filename}"
            ) from err
        except Exception as err:  # Catch-all for unexpected auth errors
            self.transport.close()
            raise exceptions.ConnectionError(
                f"Unexpected error during SSH {auth_method} auth: {err}"
            ) from err

    def disconnect(self):
        """Disconnect session."""
        if (
            hasattr(self, "transport") and self.transport and self.transport.is_active()
        ):  # Use transport directly
            self.transport.close()
        if hasattr(self, "client"):  # Ensure client is cleaned up
            self.client = None  # Or del self.client?

    def run(self, command, timeout=0):
        """Run a command on the host and return the results."""
        timeout = None if timeout == 0 else timeout
        _, stdout, stderr = self.client.exec_command(command, timeout=timeout)
        stdout.channel.recv_exit_status()  # Wait for command termination
        return helpers.Result(
            status=stdout.channel.exit_status,
            stdout=stdout.read().decode("utf-8"),
            stderr=stderr.read().decode("utf-8"),
        )

    def scp_read(self, source, destination=None, return_data=False):
        """Read a remote file into a local destination or return a string if return_data is True."""
        if not return_data:  # Ensure local destination path exists if writing to file
            destination = destination or source
            Path(destination).parent.mkdir(parents=True, exist_ok=True)

        with self.client.open_sftp() as sftp:
            if return_data:
                with sftp.open(source, "rb") as remote_file:
                    return remote_file.read().decode("utf-8")
            else:
                sftp.get(source, destination)

    def scp_write(self, source, destination=None, ensure_dir=True):
        """Write a local file to a remote destination using SCP (via SFTP)."""
        destination = destination or source
        if ensure_dir:
            self.run(
                f"mkdir -p {shlex.quote(str(Path(destination).parent))}"
            )  # Use shlex.quote for safety
        with self.client.open_sftp() as sftp:
            sftp.put(source, destination)

    def sftp_read(self, source, destination=None, return_data=False):
        """Read a remote file into a local destination or return a bytes object if return_data is True."""
        if not return_data:  # Ensure local destination path exists if writing to file
            destination = destination or source
            Path(destination).parent.mkdir(parents=True, exist_ok=True)

        with self.client.open_sftp() as sftp:
            if return_data:
                with sftp.open(source, "rb") as remote_file:
                    return remote_file.read()
            else:
                sftp.get(source, destination)

    def sftp_write(self, source, destination=None, ensure_dir=True):
        """Write a local file to a remote destination using SFTP."""
        destination = destination or source
        if ensure_dir:
            self.run(
                f"mkdir -p {shlex.quote(str(Path(destination).parent))}"
            )  # Use shlex.quote for safety
        with self.client.open_sftp() as sftp:
            sftp.put(source, destination)

    def remote_copy(self, source, dest_host, dest_path=None, ensure_dir=True):
        """Copy a file from this host to another using SFTP."""
        dest_path = dest_path or source
        if ensure_dir:  # Ensure the destination directory exists on the dest_host
            dest_host.session.run(
                f"mkdir -p {shlex.quote(str(Path(dest_path).parent))}"
            )  # Use shlex.quote for safety

        try:
            # Use file-like objects with consolidated 'with' statement
            with (
                self.client.open_sftp() as sftp_source,
                dest_host.session.client.open_sftp() as sftp_dest,
                sftp_source.open(source, "rb") as source_file,
                sftp_dest.open(dest_path, "wb") as dest_file,
            ):
                # Note: Paramiko SFTP lacks direct copy_file_to_fileobj; manually set permissions if needed.
                dest_file.set_pipelined(True)  # May improve speed
                # Read and write in chunks
                while True:
                    chunk = source_file.read(32768)  # 32KB chunks
                    if not chunk:
                        break
                    dest_file.write(chunk)
                # Attempt to copy permissions
                try:
                    stat_info = sftp_source.stat(source)
                    sftp_dest.chmod(dest_path, stat_info.st_mode)
                except Exception as e:
                    # Raise exception instead of printing
                    raise exceptions.ParamikoBindError(
                        f"Could not set permissions on remote copy {dest_path}: {e}"
                    ) from e

        except Exception as e:
            # Catch potential SFTP errors or other issues during copy
            if isinstance(e, exceptions.ParamikoBindError):
                raise  # Propagate the specific permission error
            else:
                raise exceptions.ConnectionError(f"Remote copy failed: {e}") from e

    def shell(self, pty=False):
        """Create and return an interactive shell instance."""
        channel = self.transport.open_session()  # Use transport directly
        if pty:
            channel.get_pty()  # Request PTY
        channel.invoke_shell()
        return InteractiveShell(channel)

    @contextmanager
    def tail_file(self, filename):
        """Simulate tailing a file on the remote host using SFTP stat and exec_command."""
        initial_size = -1
        try:
            with self.client.open_sftp() as sftp:
                initial_size = sftp.stat(filename).st_size
        except Exception as e:  # Handle SFTP errors (e.g., file not found)
            raise exceptions.ConnectionError(f"Could not get initial size for tail: {e}") from e

        tailer = FileTailer(initial_size=initial_size)  # Helper object to store results
        yield tailer

        # After the 'with' block, get the newly added content
        if initial_size != -1:
            # Use tail command to get content added after initial_size
            # Add 1 because tail -c +N starts from Nth byte (1-based)
            command = (
                f"tail -c +{initial_size + 1} {shlex.quote(filename)}"  # Use shlex.quote for safety
            )
            result = self.run(command)
            if result.status == 0:
                tailer.contents = result.stdout
            else:
                tailer.contents = ""  # Return empty on failure (alternative: None or raise)
        else:  # If initial stat failed
            tailer.contents = ""


class FileTailer:
    """Helper class to store results for simulated tail."""

    def __init__(self, initial_size):
        self.initial_size = initial_size
        self.contents = ""  # Content captured after context exit


class InteractiveShell:
    """A helper class that provides an interactive shell interface."""

    def __init__(self, channel):
        self._channel = channel
        self.stdout_data = ""  # Captured stdout
        self.stderr_data = ""  # Captured stderr (often merged in PTY)
        self.status = -1  # Status not reliably available via invoke_shell

    def __enter__(self):
        """Return the shell object."""
        return self

    def __exit__(self, *exc_args):
        """Close the channel and capture output."""
        # Try to capture remaining output non-blockingly
        while self._channel.recv_ready():
            self.stdout_data += self._channel.recv(1024).decode("utf-8", errors="ignore")
        while self._channel.recv_stderr_ready():
            self.stderr_data += self._channel.recv_stderr(1024).decode("utf-8", errors="ignore")

        # Exit status isn't reliably available after invoke_shell without sending 'exit'
        # and potentially blocking. Leaving status as -1.
        # self.status = self._channel.recv_exit_status()

        self._channel.close()
        # Store captured data in a Result-like object
        self.result = helpers.Result(
            status=self.status,  # Likely -1
            stdout=self.stdout_data,
            stderr=self.stderr_data,
        )

    def send(self, cmd):
        """Send a command to the channel, ensuring a newline character."""
        if not cmd.endswith("\n"):
            cmd += "\n"
        self._channel.send(cmd)
        # Avoid reading immediately after send to prevent blocking; collect output in __exit__.

    # Output is collected on exit, no separate stdout/stderr methods needed.



"""Config migrations for versions older than 0.6.9 to 0.6.9."""

from logzero import logger

TO_VERSION = "0.6.9"


def add_aap_version_setting(config_dict):
    """Add the AAP_VERSION setting to AnsibleTower provider."""
    logger.debug("Adding aap_version setting to AnsibleTower provider.")
    ansible_tower = config_dict.get("AnsibleTower", {})
    if "aap_version" not in ansible_tower:
        ansible_tower["aap_version"] = None
    config_dict["AnsibleTower"] = ansible_tower
    return config_dict


def run_migrations(config_dict):
    """Run all migrations."""
    logger.info(f"Running config migrations for {TO_VERSION}.")
    config_dict = add_aap_version_setting(config_dict)
    config_dict["_version"] = TO_VERSION
    return config_dict



"""A test provider for use in unit tests."""

import inspect

from dynaconf import Validator

from broker import helpers
from broker.providers import Provider

HOST_PROPERTIES = {
    "basic": {
        "hostname": "test.host.example.com",
        "OS": "FakeOS",
        "extra": True,
        "host_type": "host",
    }
}


class TestProvider(Provider):
    """Basic TestProvider class to test the Provider interface."""

    __test__ = False  # don't use for testing
    _validators = [Validator("TESTPROVIDER.foo", must_exist=True)]

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.config = self._settings.TESTPROVIDER.config_value
        self.foo = self._settings.TESTPROVIDER.foo

    def _host_release(self):
        caller_host = inspect.stack()[1][0].f_locals["host"]
        self.release(caller_host)

    def _set_attributes(self, host_inst, broker_args=None):
        host_inst.__dict__.update(
            {
                "release": self._host_release,
                "_prov_inst": self,
                "_broker_provider": "TestProvider",
                "_broker_provider_instance": self.instance,
                "_broker_args": broker_args,
            }
        )

    def construct_host(self, provider_params, host_classes, **kwargs):
        """Construct a host object from the provider_params and kwargs."""
        if provider_params:
            host_params = provider_params.copy()
            host_params.update(kwargs)
            host_inst = host_classes[host_params["host_type"]](
                broker_settings=self._settings, **host_params
            )
        else:  # if we are reconstructing the host from the inventory
            host_inst = host_classes[kwargs.get("type", "host")](**kwargs)
        self._set_attributes(host_inst, broker_args=kwargs)
        return host_inst

    @Provider.register_action()
    def test_action(self, **kwargs):
        """A dummy action for testing."""
        action = kwargs.get("test_action")
        if action == "release":
            return "released", kwargs
        if action in HOST_PROPERTIES:
            return HOST_PROPERTIES
        return HOST_PROPERTIES["basic"]

    def release(self, host_obj):
        """Release a host ;) ."""
        return self.test_action(test_action="release", **host_obj.to_dict())

    def extend(self):
        """No current implementation for this provider."""

    def get_inventory(self, *args, **kwargs):
        """Load a filtered local inventory."""
        return helpers.load_inventory(
            filter=f'@inv._broker_provider == "{self.__class__.__name__}"'
        )

    def provider_help(self):
        """No current implementation for this provider."""



"""Module providing the functionality powering the `broker config` command."""

import importlib
from importlib.metadata import version
import json
import os
from pathlib import Path
import pkgutil
import sys

import click
from logzero import logger
from packaging.version import Version
from ruamel.yaml import YAML, YAMLError

from broker import exceptions

yaml = YAML()
yaml.default_flow_style = False
yaml.sort_keys = False

C_SEP = "."  # chunk separator
GH_CFG = "https://raw.githubusercontent.com/SatelliteQE/broker/master/broker_settings.yaml.example"


def file_name_to_ver(file_name):
    """Convert a version-encoded filename `v0_6_0` to a `Version` object."""
    return Version(file_name[1:].replace("_", "."))


class ConfigManager:
    """Class to interact with Broker's configuration file.

    One important concept of these commands is the concept of a "chunk".

    A chunk is a part of the configuration file that can be accessed or updated.
    Chunks are specified by their keys in the configuration file.
    Nested chunks are separated by periods.

    e.g. broker config view AnsibleTower.instances.my_instance
    """

    interactive_mode = sys.stdin.isatty()
    version = version("broker")
    no_global_config = False

    def __init__(self, settings_path=None):
        self._settings_path = settings_path
        if settings_path:
            if settings_path.exists():
                self._cfg = yaml.load(self._settings_path)
            elif self.no_global_config or os.environ.get("BROKER_NO_GLOBAL_CONFIG"):
                logger.info("Continuing without global config file.")
                return
            else:
                click.secho(
                    f"Broker settings file not found at {settings_path.absolute()}.", fg="red"
                )
                self.init_config_file()

    def _interactive_edit(self, chunk):
        """Write the chunk data to a temporary file and open it in an editor."""
        temp_file = Path("temp_settings.yaml")
        yaml.dump(chunk, temp_file)
        click.edit(filename=str(temp_file))
        new_data = temp_file.read_text()
        temp_file.unlink()
        # first try to load it as yaml
        try:
            return yaml.load(new_data)
        except YAMLError:  # then try json
            try:
                return json.loads(new_data)
            except json.JSONDecodeError:  # finally, just return the raw data
                return new_data

    def _import_config(self, source, is_url=False):
        """Initialize the broker settings file from a source."""
        proceed = True
        if self.interactive_mode:
            try:
                proceed = click.confirm(f"Get example file from {source}?")
            except click.core.Abort:
                # We're likely in a different non-interactive environment (container?)
                self._interactive_mode = False
        if not proceed:
            return
        # get example file from source
        if is_url:
            import requests

            click.echo(f"Downloading example file from: {source}")
            return requests.get(source, timeout=60, verify=False).text
        else:
            return source.read_text()

    def _get_migrations(self, force_version=None):
        """Construct a list of all applicable migrations."""
        from broker import config_migrations

        config_version = Version(self._cfg.get("_version", "0.0.0"))
        if force_version:
            force_version = Version(force_version)
        migrations = []
        for _, name, _ in pkgutil.iter_modules(config_migrations.__path__):
            module = importlib.import_module(f"broker.config_migrations.{name}")
            if hasattr(module, "run_migrations") and "example" not in module.__name__:
                if force_version and force_version == file_name_to_ver(name):
                    migrations.append(module)
                    break
                elif config_version < file_name_to_ver(name):
                    migrations.append(module)
        return migrations

    def backup(self):
        """Backup the current configuration file."""
        logger.debug(
            f"Backing up the configuration file to {self._settings_path.with_suffix('.bak')}"
        )
        self._settings_path.with_suffix(".bak").write_text(self._settings_path.read_text())

    def restore(self):
        """Restore the configuration file from a backup if it exists."""
        logger.debug(
            f"Restoring the configuration file from {self._settings_path.with_suffix('.bak')}"
        )
        backup_path = self._settings_path.with_suffix(".bak")
        if not backup_path.exists():
            raise exceptions.UserError("No backup file found.")
        self._settings_path.write_text(backup_path.read_text())

    def edit(self, chunk=None, content=None):
        """Open the config file in an editor."""
        if not self.interactive_mode:
            raise exceptions.UserError(
                "Attempted to edit the config in non-interactive mode.\n"
                "Did you mean to use the `set` method instead?"
            )
        content = content or self.get(chunk=chunk)
        new_val = self._interactive_edit(content)
        self.update(chunk, new_val)

    def get(self, chunk=None, curr_chunk=None, suppress=False):
        """Get a chunk of Broker's config or the whole config."""
        if not curr_chunk:
            curr_chunk = self._cfg
        if not chunk:
            return curr_chunk
        if C_SEP in chunk:
            curr, chunk = chunk.split(C_SEP, 1)
            # curr = int(curr) if curr.isdigit() else curr
            return self.get(chunk, curr_chunk=curr_chunk[curr])
        else:
            # chunk = int(chunk) if chunk.isdigit() else chunk
            try:
                return curr_chunk[chunk]
            except KeyError:
                if suppress:
                    return
                raise exceptions.UserError(f"Chunk '{chunk}' not found in the config.")

    def update(self, chunk, new_val, curr_chunk=None):
        """Update a chunk of Broker's config or the whole config."""
        # Recursive down to find the chunk to update, then propagate the new value back up
        if not curr_chunk:  # we're at the top level, so update the config directly
            if chunk is None:  # the whole config is being updated
                self._cfg = new_val
            elif C_SEP in chunk:  # the update needs to happen at a lower level
                curr, chunk = chunk.split(C_SEP, 1)
                self._cfg[curr] = self.update(chunk, new_val, curr_chunk=self._cfg[curr])
            else:
                self._cfg[chunk] = new_val
            # update the config file if it exists
            if self._settings_path.exists():
                self.backup()
            yaml.dump(self._cfg, self._settings_path)
        else:  # we're not at the top level, so keep going down
            if C_SEP in chunk:
                curr, chunk = chunk.split(C_SEP, 1)
                curr_chunk[curr] = self.update(chunk, new_val, curr_chunk=curr_chunk[curr])
            else:
                curr_chunk[chunk] = new_val
            return curr_chunk

    def nicks(self, nick=None):
        """Get a list of nicks or single nick information."""
        nicks = self.get("nicks")
        if nick:
            return nicks[nick]
        return list(nicks.keys())

    def init_config_file(self, chunk=None, _from=None):
        """Check for the existence of the config file and create it if it doesn't exist."""
        if self.interactive_mode and self._settings_path.exists() and not chunk:
            # if the file exists, ask the user if they want to overwrite it
            if (
                click.prompt(
                    f"Overwrite the settings file at {self._settings_path.absolute()}. Overwrite?",
                    type=click.Choice(["y", "n"]),
                    default="n",
                )
                != "y"
            ):
                return
        raw_data = None
        if _from:
            # determine if this is a local file or a URL
            if Path(_from).exists():
                raw_data = self._import_config(Path(_from))
            else:
                raw_data = self._import_config(_from, is_url=True)
        # if we still don't have data, get the example file from the local repo or GitHub
        if not raw_data:
            # get the example file from the local repo or GitHub
            example_path = Path(__file__).parent.parent.joinpath("broker_settings.yaml.example")
            if example_path.exists():
                raw_data = self._import_config(example_path)
            if not raw_data:
                raw_data = self._import_config(GH_CFG, is_url=True)
        if not raw_data:
            raise exceptions.ConfigurationError(
                f"Broker settings file not found at {self._settings_path.absolute()}."
                "\nIf you want to disable the global config file, "
                "set a BROKER_NO_GLOBAL_CONFIG environment variable"
            )
        chunk_data = self.get(chunk, yaml.load(raw_data))
        if self.interactive_mode:
            chunk_data = self._interactive_edit(chunk_data)
        self.update(chunk, chunk_data)

    def migrate(self, force_version=None):
        """Migrate the config from a previous version of Broker."""
        # get all available migrations
        if not (migrations := self._get_migrations(force_version)):
            logger.info("No migrations are applicable to your config.")
            return
        # run all migrations in order
        working_config = self._cfg
        for migration in sorted(migrations, key=lambda m: m.TO_VERSION):
            working_config = migration.run_migrations(working_config)
        self.backup()
        yaml.dump(working_config, self._settings_path)
        logger.info("Config migration complete.")

    def validate(self, chunk, providers):
        """Validate a top-level chunk of Broker's config."""
        if chunk == "all":
            all_settings = [prov for prov in providers if prov != "TestProvider"] + ["base", "ssh"]
            for item in all_settings:
                self.validate(item, providers)
            return
        chunk = chunk.split(C_SEP)[0] if C_SEP in chunk else chunk
        if chunk.lower() == "base":
            return
        if chunk.lower() == "ssh":
            from broker.settings import settings

            logger.info("Validating SSH settings.")
            settings.validators.validate(only="SSH")
            return
        instance_settings = {}
        if ":" in chunk:
            chunk, instance = chunk.split(":")
            instance_settings = {chunk: instance}
        if chunk not in providers:
            raise exceptions.UserError(
                "I don't know how to validate that.\n"
                "If it's important, it is likely covered in the base validations."
            )
        if not self.get(chunk, suppress=True):
            logger.warning(f"No settings found for {chunk} provider.")
            return
        logger.info(f"Validating {chunk} provider settings.")
        try:
            providers[chunk](**instance_settings)
        except Exception as err:  # noqa: BLE001
            logger.warning(f"Provider {chunk} failed validation: {err}")



"""Defines the CLI commands for Broker."""

from functools import wraps
import signal
import sys

from logzero import logger
from rich.console import Console
from rich.syntax import Syntax
from rich.table import Table
import rich_click as click

from broker import exceptions, helpers, settings
from broker.broker import Broker
from broker.config_manager import ConfigManager
from broker.logger import LOG_LEVEL
from broker.providers import PROVIDER_ACTIONS, PROVIDER_HELP, PROVIDERS

signal.signal(signal.SIGINT, helpers.handle_keyboardinterrupt)
CONSOLE = Console(no_color=settings.settings.less_colors)  # rich console for pretty printing

click.rich_click.SHOW_ARGUMENTS = True
click.rich_click.COMMAND_GROUPS = {
    "broker": [
        {"name": "Core Actions", "commands": ["checkout", "checkin", "inventory"]},
        {"name": "Extras", "commands": ["execute", "extend", "providers", "config"]},
    ]
}


def guarded_command(group=None, *cli_args, **cli_kwargs):
    """Wrap commands with logging and exception handling."""
    if not group:
        group = cli  # default to the main cli group

    def decorator(func):
        @group.command(*cli_args, **cli_kwargs)
        @wraps(func)
        def wrapper(*args, **kwargs):
            try:
                logger.log(LOG_LEVEL.TRACE.value, f"Calling {func=}(*{args=} **{kwargs=})")
                retval = func(*args, **kwargs)
                logger.log(
                    LOG_LEVEL.TRACE.value,
                    f"Finished {func=}(*{args=} **{kwargs=}) {retval=}",
                )
                helpers.emit(return_code=0)
                return retval
            except Exception as err:  # noqa: BLE001 -- we want to catch all exceptions
                if not isinstance(err, exceptions.BrokerError):
                    err = exceptions.BrokerError(err)
                    logger.error(f"Command failed: {err.message}")
                else:  # BrokerError children already log their messages
                    logger.error(f"Command failed due to: {type(err).__name__}")
                helpers.emit(return_code=err.error_code, error_message=str(err.message))
                sys.exit(err.error_code)

        return wrapper

    return decorator


def parse_labels(provider_labels):
    """Parse the provided label string and returns labels in a dict."""
    return {
        label[0]: "=".join(label[1:])
        for label in [kv_pair.split("=") for kv_pair in provider_labels.split(",")]
    }


def provider_options(command):
    """Apply provider-specific decorators to each command this decorates."""
    for prov in PROVIDERS.values():
        if not settings.settings.get(prov.__name__):
            continue
        for option in getattr(prov, f"_{command.__name__}_options"):
            command = option(command)
    return command


def populate_providers(click_group):
    """Populate the subcommands for providers subcommand using provider information.

    Providers become subcommands and their actions become arguments to their subcommand.

    Example:
        Usage: broker providers AnsibleTower [OPTIONS]

        Options:
        --workflows      Get available workflows
        --workflow TEXT  Get information about a workflow
        --help           Show this message and exit.

    Note: This currently only works for the default instance for each provider
    """
    for prov, prov_class in (pairs for pairs in PROVIDERS.items()):
        if not settings.settings.get(prov_class.__name__):
            continue

        @guarded_command(
            group=click_group,
            name=prov,
            context_settings={
                "allow_extra_args": True,
                "ignore_unknown_options": True,
            },
        )
        @click.pass_context
        def provider_cmd(ctx, *args, **kwargs):  # the actual subcommand
            """Get information about a provider's actions."""
            # add additional args flags to the kwargs
            for arg in ctx.args:
                if arg.startswith("--"):
                    kwargs[arg[2:]] = True
            # if additional arguments were passed, include them in the broker args
            # strip leading -- characters
            kwargs.update(
                {
                    (key[2:] if key.startswith("--") else key): val
                    for key, val in zip(ctx.args[::2], ctx.args[1::2])
                }
            )
            broker_inst = Broker(**kwargs)
            broker_inst.provider_help(ctx.info_name)

        # iterate through available actions and populate options from them
        for option, p_cls, is_flag, alt_text in PROVIDER_HELP:
            if p_cls is not prov_class:
                continue
            option = option.replace("_", "-")  # noqa: PLW2901
            if is_flag:
                provider_cmd = click.option(
                    f"--{option}", is_flag=True, help=alt_text or f"Get available {option}"
                )(provider_cmd)
            else:
                provider_cmd = click.option(
                    f"--{option}", type=str, help=alt_text or f"Get information about a {option}"
                )(provider_cmd)
        provider_cmd = click.option(
            "--results-limit",
            type=int,
            help="The maximum number of results to get back",
        )(provider_cmd)
        provider_cmd = click.option(
            "--results-filter",
            type=str,
            help="Apply a broker filter to returned results",
        )(provider_cmd)


@click.group(invoke_without_command=True)
@click.option(
    "--log-level",
    type=click.Choice(["info", "warning", "error", "debug", "trace", "silent"]),
    default=settings.settings.logging.console_level,
    callback=helpers.update_log_level,
    is_eager=True,
    expose_value=False,
)
@click.option(
    "--output-file",
    type=click.Path(dir_okay=False),
    callback=helpers.set_emit_file,
    is_eager=True,
    expose_value=False,
    help="Path to file where emitted json values should be stored",
)
@click.option(
    "--version",
    is_flag=True,
    help="Get broker system-level information",
)
def cli(version):
    """Command-line interface for interacting with providers."""
    if version:
        from packaging.version import Version
        import requests

        # Check against the latest version published to PyPi
        try:
            latest_version = Version(
                requests.get("https://pypi.org/pypi/broker/json", timeout=60).json()["info"][
                    "version"
                ]
            )
            if latest_version > Version(ConfigManager.version):
                click.secho(
                    f"A newer version of broker is available: {latest_version}",
                    fg="yellow",
                )
        except requests.exceptions.RequestException as err:
            logger.warning(f"Unable to check for latest version: {err}")

        # Create a rich table
        table = Table(title=f"Broker {ConfigManager.version}")

        table.add_column("", justify="left", style="cyan", no_wrap=True)
        table.add_column("Location", justify="left", style="magenta")

        table.add_row("Broker Directory", str(settings.BROKER_DIRECTORY.absolute()))
        table.add_row("Settings File", str(settings.settings_path.absolute()))
        table.add_row("Inventory File", f"{settings.BROKER_DIRECTORY.absolute()}/inventory.yaml")
        table.add_row("Log File", f"{settings.BROKER_DIRECTORY.absolute()}/logs/broker.log")

        # Print the table
        CONSOLE.print(table)


@guarded_command(context_settings={"allow_extra_args": True, "ignore_unknown_options": True})
@click.option("-b", "--background", is_flag=True, help="Run checkout in the background")
@click.option("-n", "--nick", type=str, help="Use a nickname defined in your settings")
@click.option("-c", "--count", type=int, help="Number of times broker repeats the checkout")
@click.option(
    "-l",
    "--provider-labels",
    type=str,
    help="A string representing the list"
    " of k=v pairs (comma-separated) to be used as provider resource"
    " labels (e.g. '-l k1=v1,k2=v2,k3=v3=z4').",
)
@click.option(
    "--args-file",
    type=click.Path(exists=True),
    help="A json or yaml file mapping arguments to values",
)
@provider_options
@click.pass_context
def checkout(ctx, background, nick, count, args_file, provider_labels, **kwargs):
    """Checkout or "create" a Virtual Machine broker instance.

    COMMAND: broker checkout --workflow "workflow-name" --workflow_arg1 something

    COMMAND: broker checkout --nick "nickname"

    You can also checkout against a non-default provider instance, e.g:

    COMMAND: broker checkout ... --AnsibleTower <instance name>
    """
    broker_args = helpers.clean_dict(kwargs)
    if nick:
        broker_args["nick"] = nick
    if count:
        broker_args["_count"] = count
    if args_file:
        broker_args["args_file"] = args_file
    if provider_labels:
        broker_args["provider_labels"] = parse_labels(provider_labels)

    broker_args.update(helpers.kwargs_from_click_ctx(ctx))
    if background:
        helpers.fork_broker()
    Broker(**broker_args).checkout()


@cli.group()
def providers():
    """Get information about a provider and its actions."""
    pass


populate_providers(providers)


@guarded_command()
@click.argument("hosts", type=str, nargs=-1)
@click.option("-b", "--background", is_flag=True, help="Run checkin in the background")
@click.option("--all", "all_", is_flag=True, help="Select all hosts")
@click.option("--sequential", is_flag=True, help="Run checkins sequentially")
@click.option("--filter", type=str, help="Checkin only what matches the specified filter")
def checkin(hosts, background, all_, sequential, filter):
    """Checkin or "remove" a host or series of hosts.

    COMMAND: broker checkin <hostname>|<local id>|--all
    """
    if background:
        helpers.fork_broker()
    inventory = helpers.load_inventory(filter=filter)
    to_remove = []
    unmatched = set(hosts)  # Track unmatched hosts

    for num, host in enumerate(inventory):
        # Check if this host matches any of our criteria
        if (
            all_
            or str(num) in unmatched
            or host.get("hostname") in unmatched
            or host.get("name") in unmatched
        ):
            to_remove.append(Broker().reconstruct_host(host))
            # Remove all possible match values for this host from unmatched
            unmatched.discard(str(num))
            unmatched.discard(host.get("hostname"))
            unmatched.discard(host.get("name"))

    if unmatched:
        logger.warning(f"The following hosts were not found in inventory: {', '.join(unmatched)}")
    if to_remove:
        Broker(hosts=to_remove).checkin(sequential=sequential)


@guarded_command()
@click.option("--details", is_flag=True, help="Display all host details")
@click.option("--list", "_list", is_flag=True, help="Display only hostnames and local ids")
@click.option(
    "--sync",
    type=str,
    help="Class-style name of a supported broker provider. (AnsibleTower)",
)
@click.option("--filter", type=str, help="Display only what matches the specified filter")
def inventory(details, _list, sync, filter):
    """Display a table of hosts in your local inventory.

    Inventory fields are configurable in Broker's settings file.

    Run a sync for your providers to pull down your host information.

    e.g. `broker inventory --sync AnsibleTower`

    Note: Applying a filter will result in incorrect id's being displayed.
    """
    if sync:
        Broker.sync_inventory(provider=sync)
    inventory = helpers.load_inventory(filter=filter)
    helpers.emit({"inventory": inventory})
    # details is handled differently than the normal and list views
    if details:
        detailed = helpers.yaml_format(dict(enumerate(inventory)))
        CONSOLE.print(Syntax(detailed, "yaml", background_color="default"))
        return

    inventory_fields = (
        {"Host": settings.settings.inventory_list_vars}
        if _list
        else settings.settings.inventory_fields
    )
    curated_host_info = [
        helpers.inventory_fields_to_dict(
            inventory_fields=inventory_fields,
            host_dict=host,
            provider_actions=PROVIDER_ACTIONS,
        )
        for host in inventory
    ]
    if not curated_host_info:
        CONSOLE.print("No hosts found in inventory.")
        return
    table = helpers.dictlist_to_table(curated_host_info, "Host Inventory", _id=True)
    if _list:
        table.title = None
        table.box = None
    CONSOLE.print(table)


@guarded_command()
@click.argument("hosts", type=str, nargs=-1)
@click.option("-b", "--background", is_flag=True, help="Run extend in the background")
@click.option("--all", "all_", is_flag=True, help="Select all hosts")
@click.option("--sequential", is_flag=True, help="Run extends sequentially")
@click.option("--filter", type=str, help="Extend only what matches the specified filter")
@click.option(
    "-l",
    "--provider-labels",
    type=str,
    help="A string representing the list"
    " of k=v pairs (comma-separated) to be used as provider resource"
    " labels (e.g. '-l k1=v1,k2=v2,k3=v3=z4').",
)
@provider_options
def extend(hosts, background, all_, sequential, filter, **kwargs):
    """Extend a host's lease time.

    COMMAND: broker extend <hostname>|<host name>|<local id>|--all
    """
    broker_args = helpers.clean_dict(kwargs)
    if background:
        helpers.fork_broker()
    inventory = helpers.load_inventory(filter=filter)
    to_extend = []
    for num, host in enumerate(inventory):
        if str(num) in hosts or host["hostname"] in hosts or host.get("name") in hosts or all_:
            to_extend.append(Broker().reconstruct_host(host))
    Broker(hosts=to_extend, **broker_args).extend(sequential=sequential)


@guarded_command(context_settings={"allow_extra_args": True, "ignore_unknown_options": True})
@click.option("-b", "--background", is_flag=True, help="Run execute in the background")
@click.option("--nick", type=str, help="Use a nickname defined in your settings")
@click.option("--output-format", "-o", type=click.Choice(["log", "raw", "yaml"]), default="log")
@click.option(
    "--artifacts",
    type=click.Choice(["merge", "last"]),
    help="AnsibleTower: return artifacts associated with the execution.",
)
@click.option(
    "--args-file",
    type=click.Path(exists=True),
    help="A json or yaml file mapping arguments to values",
)
@click.option(
    "-l",
    "--provider-labels",
    type=str,
    help="A string representing the list"
    " of k=v pairs (comma-separated) to be used as provider resource"
    " labels (e.g. '-l k1=v1,k2=v2,k3=v3=z4').",
)
@provider_options
@click.pass_context
def execute(ctx, background, nick, output_format, artifacts, args_file, provider_labels, **kwargs):
    """Execute an arbitrary provider action.

    COMMAND: broker execute --workflow "workflow-name" --workflow_arg1 something

    COMMAND: broker execute --nick "nickname"
    """
    broker_args = helpers.clean_dict(kwargs)
    if nick:
        broker_args["nick"] = nick
    if artifacts:
        broker_args["artifacts"] = artifacts
    if args_file:
        broker_args["args_file"] = args_file
    if provider_labels:
        broker_args["provider_labels"] = parse_labels(provider_labels)

    broker_args.update(helpers.kwargs_from_click_ctx(ctx))

    if background:
        helpers.fork_broker()
    result = Broker(**broker_args).execute()
    helpers.emit({"output": result})
    if output_format == "raw":
        click.echo(result)
    elif output_format == "log":
        logger.info(result)
    elif output_format == "yaml":
        click.echo(helpers.yaml_format(result))


@cli.group()
def config():
    """View and manage Broker's configuration.

    Note: One important concept of these commands is the concept of a "chunk".

    A chunk is a part of the configuration file that can be accessed or updated.
    Chunks are specified by their keys in the configuration file.
    Nested chunks are separated by periods.

    e.g. broker config view AnsibleTower.instances.my_instance
    """


@guarded_command(group=config)
@click.argument("chunk", type=str, required=False)
@click.option("--no-syntax", is_flag=True, help="Disable syntax highlighting")
def view(chunk, no_syntax):
    """View all or part of the broker configuration."""
    result = helpers.yaml_format(ConfigManager(settings.settings_path).get(chunk))
    if no_syntax:
        CONSOLE.print(result)
    else:
        CONSOLE.print(Syntax(result, "yaml", background_color="default"))


@guarded_command(group=config)
@click.argument("chunk", type=str, required=False)
def edit(chunk):
    """Directly edit the broker configuration file.

    You can define the scope of the edit by specifying a chunk.
    Otherwise, the entire configuration file will be opened.
    """
    ConfigManager(settings.settings_path).edit(chunk)


@guarded_command(group=config, name="set")
@click.argument("chunk", type=str, required=True)
@click.argument("new-value", type=str, required=True)
def _set(chunk, new_value):
    """Set a value in the Broker configuration file.

    These updates take the form of `<chunk> <value>` pairs.
    You can also pass a yaml or json file containing the new contents of a chunk.
    """
    new_value = helpers.resolve_file_args({"nv": new_value})["nv"]
    ConfigManager(settings.settings_path).update(chunk, new_value)


@guarded_command(group=config)
def restore():
    """Restore the broker configuration file to the last backup."""
    ConfigManager(settings.settings_path).restore()


@guarded_command(group=config)
@click.argument("chunk", type=str, required=False)
@click.option("--from", "_from", type=str, help="A file path or URL to initialize the config from.")
def init(chunk=None, _from=None):
    """Initialize the broker configuration file from your local clone or GitHub.

    You can also init specific chunks by passing the chunk name.
    Additionally, if you want to initialize from a file or URL, you can pass the `--from` flag.
    Keep in mind that the file and url contents need to be valid yaml.
    """
    ConfigManager(settings.settings_path).init_config_file(chunk=chunk, _from=_from)


@guarded_command(group=config)
def nicks():
    """Get a list of nicks."""
    result = ConfigManager(settings.settings_path).nicks()
    CONSOLE.print("\n".join(result))


@guarded_command(group=config)
@click.argument("nick", type=str, required=True)
@click.option("--no-syntax", is_flag=True, help="Disable syntax highlighting")
def nick(nick, no_syntax):
    """Get information about a specific nick."""
    result = helpers.yaml_format(ConfigManager(settings.settings_path).nicks(nick))
    if no_syntax:
        CONSOLE.print(result)
    else:
        CONSOLE.print(Syntax(result, "yaml", background_color="default"))


@guarded_command(group=config)
@click.option("-f", "--force-version", type=str, help="Force the migration to a specific version")
def migrate(force_version=None):
    """Migrate the broker configuration file to the latest version."""
    ConfigManager(settings.settings_path).migrate(force_version=force_version)


@guarded_command(group=config)
@click.argument("chunk", type=str, required=False, default="base")
def validate(chunk):
    """Validate top-level chunks of the broker configuration file.

    You can validate against the `base` settings by default or specify a provider.
    You can also validate against a specific provider instance with `ProviderClass:instance_name`.

    To validate everything, pass `all`
    """
    try:
        ConfigManager(settings.settings_path).validate(chunk, PROVIDERS)
        logger.info("Validation passed!")
    except exceptions.BrokerError as err:
        logger.warning(f"Validation failed: {err}")



"""Main interface for the Broker API.

This module provides the main interface for the Broker API, which allows users to
manage cloud resources across multiple providers.

It defines the `Host` class, which represents a cloud resource, and the `Broker` class,
which provides methods for managing hosts.

The `Host` class is defined in the `broker.hosts` module,
and the provider classes are defined in the `broker.providers` module.

Exceptions are defined in the `broker.exceptions` module,
and helper functions are defined in the `broker.helpers` module.

Note:
    This module (or parent directory) should be used as the main entry point for the Broker API.

"""

from concurrent.futures import ThreadPoolExecutor, as_completed
from contextlib import contextmanager

from logzero import logger

from broker import exceptions, helpers, logger as broker_logger
from broker.hosts import Host
from broker.providers import PROVIDER_ACTIONS, PROVIDERS, _provider_imports
from broker.settings import clone_global_settings

# load all the provider class so they are registered
for _import in _provider_imports:
    __import__(f"broker.providers.{_import}", globals(), locals(), [], 0)


def _try_teardown(host_obj):
    """Try a host's teardown method and return an exception message if needed."""
    try:
        host_obj.teardown()
    except Exception as err:  # noqa: BLE001
        logger.debug(f"Tell Jake the exception was: {err}")
        return exceptions.HostError(host_obj, f"error during teardown:\n{err}")


class Broker:
    """Main Broker class to be used as the primary interface for the Broker API."""

    # map exceptions for easier access when used as a library
    AuthenticationError = exceptions.AuthenticationError
    BrokerError = exceptions.BrokerError
    ConfigurationError = exceptions.ConfigurationError
    NotImplementedError = exceptions.NotImplementedError
    PermissionError = exceptions.PermissionError
    ProviderError = exceptions.ProviderError
    UserError = exceptions.UserError

    def __init__(self, broker_settings=None, **kwargs):
        self._settings = broker_settings or clone_global_settings()
        if broker_settings:
            logger.debug(f"Using local settings object: {self._settings.to_dict()}")
            if "logging" in broker_settings:
                broker_logger.setup_logzero(
                    level=broker_settings.logging.console_level,
                    formatter=kwargs.pop("broker_log_formatter", None),
                    file_level=broker_settings.logging.file_level,
                    path=broker_settings.logging.get("file_path"),
                )
        else:
            logger.debug("Using global settings.")
        kwargs = helpers.resolve_file_args(kwargs)
        logger.debug(f"Broker instantiated with {kwargs=}")
        self._hosts = kwargs.pop("hosts", [])
        self.host_classes = {"host": Host}
        # if a nick was specified, pull in the resolved arguments
        if "nick" in kwargs:
            nick = kwargs.pop("nick")
            kwargs = helpers.merge_dicts(kwargs, helpers.resolve_nick(nick, self._settings))
            logger.debug(f"kwargs after nick resolution {kwargs=}")
        # Allow users to more simply pass a host class instead of a dict
        if "host_class" in kwargs:
            self.host_classes["host"] = kwargs.pop("host_class")
        elif "host_classes" in kwargs:
            self.host_classes.update(kwargs.pop("host_classes"))
        # determine the provider actions based on kwarg parameters
        self._provider_actions = {}
        self._update_provider_actions(kwargs)
        self._kwargs = kwargs

    def _act(self, provider, method, checkout=False):
        """Perform a general action against a provider's method."""
        count = self._kwargs.get("_count", 1)
        logger.debug(f"Resolving action {method} on provider {provider}.")
        provider_inst = provider(broker_settings=self._settings, **self._kwargs)
        helpers.emit(
            {
                "provider": provider_inst.__class__.__name__,
                "action": method,
                "arguments": self._kwargs,
            }
        )
        method_obj = getattr(provider_inst, method)
        logger.debug(f"On {provider_inst=} executing {method_obj=} with params {self._kwargs=}.")
        # Overkill for a single action, cleaner than splitting the logic
        max_workers = (
            min(count, int(self._settings.thread_limit)) if self._settings.thread_limit else None
        )
        with ThreadPoolExecutor(max_workers=max_workers) as workers:
            tasks = [workers.submit(method_obj, **self._kwargs) for _ in range(count)]
            result = []
            for task in as_completed(tasks):
                try:
                    result.append(task.result())
                except exceptions.ProviderError as err:  # noqa: PERF203
                    result.append(err)
        logger.debug(f"Result:\n{result}")
        if result and checkout:
            return [
                provider_inst.construct_host(
                    provider_params=res, host_classes=self.host_classes, **self._kwargs
                )
                if not isinstance(res, exceptions.ProviderError)
                else res
                for res in result
            ]
        else:
            return result if len(result) > 1 else result[0]

    def _update_provider_actions(self, kwargs):
        if not self._provider_actions:
            for key, action in PROVIDER_ACTIONS.items():
                if key in kwargs:
                    self._provider_actions[key] = action

    def checkout(self):
        """Checkout one or more hosts.

        :return: Host obj or list of Host objects
        """
        logger.debug(f"Starting checkout(s): {self._provider_actions=}")
        if not self._provider_actions:
            raise self.BrokerError("Could not determine an appropriate provider")
        for action in self._provider_actions:
            provider, method = PROVIDER_ACTIONS[action]
            logger.info(f"Using provider {provider.__name__} to checkout")
        # perform the checkout, concurrently if _count is set
        hosts = self._act(provider, method, checkout=True)
        err = None
        for host in hosts[:]:
            if isinstance(host, exceptions.ProviderError):
                err = host
                hosts.remove(host)
            else:
                logger.info(f"{host.__class__.__name__}: {host.hostname}")
        helpers.emit(hosts=[host.to_dict() for host in hosts])
        self._hosts.extend(hosts)
        helpers.update_inventory([host.to_dict() for host in hosts])
        if isinstance(err, exceptions.BrokerError):
            raise err  # Let our own error raise cleanly
        elif err:
            raise self.BrokerError(f"Error during checkout from {self}") from err
        return hosts if len(hosts) != 1 else hosts[0]

    def execute(self, **kwargs):
        """Execute a provider action.

        :return: Any results given back by the provider
        """
        self._update_provider_actions(kwargs)
        self._kwargs.update(kwargs)
        if not self._provider_actions:
            raise self.BrokerError("Could not determine an appropriate provider")
        for action in self._provider_actions:
            provider, method = PROVIDER_ACTIONS[action]
        logger.info(f"Using provider {provider.__name__} for execution")
        return self._act(provider, method)

    def provider_help(self, provider_name):
        """Use a provider's provider_help method to get argument information."""
        provider = PROVIDERS[provider_name]
        logger.info(f"Querying provider {provider.__name__}")
        self._act(provider, "provider_help", checkout=False)

    def _checkin(self, host):
        logger.info(f"Checking in {host.hostname or host.name}")
        host.close()
        try:
            host.release()
        except Exception as err:
            logger.warning(f"Encountered exception during checkin: {err}")
            raise
            # pass
        return host

    def checkin(self, sequential=False, host=None, in_context=False):
        """Checkin one or more VMs.

        :param host: can be one of:
            None - Will use the contents of self._hosts
            A single host object
            A list of host objects
            A dictionary mapping host types to one or more host objects

        :param sequential: boolean whether to run checkins sequentially
        :param in_context: Whether checkin is part of context manager
        """
        # default to hosts listed on the instance
        hosts = host or self._hosts
        logger.debug(
            f"Checkin called with: {hosts}, running {'sequential' if sequential else 'concurrent'}"
        )
        # normalize the type since the function accepts multiple types
        if isinstance(hosts, dict):
            # flatten the lists of hosts from the values of the dict
            hosts = [host for host_list in hosts.values() for host in host_list]
        else:
            if not isinstance(hosts, list):
                hosts = [hosts]
            if in_context:
                hosts = [
                    host for host in hosts if not getattr(host, "_skip_context_checkin", False)
                ]
        if not hosts:
            logger.debug("Checkin called with no hosts, taking no action")
            return
        max_workers = (
            min(len(hosts), int(self._settings.thread_limit))
            if self._settings.thread_limit
            else None
        )
        with ThreadPoolExecutor(max_workers=1 if sequential else max_workers) as workers:
            completed_checkins = as_completed(
                # reversing over a copy of the list to avoid skipping
                workers.submit(self._checkin, _host)
                for _host in hosts[::-1]
            )
            for completed in completed_checkins:
                _host = completed.result()
                self._hosts = [h for h in self._hosts if h.to_dict() != _host.to_dict()]
                logger.debug(f"Completed checkin process for {_host.hostname or _host.name}")
        helpers.update_inventory(remove=[h.name for h in hosts])

    def _extend(self, host):
        """Extend a single VM."""
        logger.info(f"Extending host {host.hostname}")
        provider = PROVIDERS[host._broker_provider]
        self._kwargs["target_vm"] = host
        self._act(provider, "extend", checkout=False)
        return host

    def extend(self, sequential=False, host=None):
        """Extend one or more VMs.

        :param host: can be one of:
            None - Will use the contents of self._hosts
            A single host object
            A list of host objects
            A dictionary mapping host types to one or more host objects

        :param sequential: boolean whether to run checkins sequentially
        """
        # default to hosts listed on the instance
        hosts = host or self._hosts
        logger.debug(
            f"Extend called with: {hosts}, running {'sequential' if sequential else 'concurrent'}"
        )
        # normalize the type since the function accepts multiple types
        if isinstance(hosts, dict):
            # flatten the lists of hosts from the values of the dict
            hosts = [host for host_list in hosts.values() for host in host_list]
        if not isinstance(hosts, list):
            hosts = [hosts]

        if not hosts:
            logger.debug("Extend called with no hosts, taking no action")
            return

        with ThreadPoolExecutor(max_workers=1 if sequential else len(hosts)) as workers:
            completed_extends = as_completed(workers.submit(self._extend, _host) for _host in hosts)
            for completed in completed_extends:
                _host = completed.result()
                logger.info(f"Completed extend for {_host.hostname or _host.name}")

    @staticmethod
    def sync_inventory(provider, broker_settings=None):
        """Acquire a list of hosts from a provider and update our inventory."""
        _settings = broker_settings or clone_global_settings()
        additional_arg, instance = None, {}
        if "::" in provider:
            provider, instance = provider.split("::")
        if ":" in provider:
            provider, additional_arg = provider.split(":")
        logger.info(f"Pulling remote inventory from {f'{instance} ' if instance else ''}{provider}")
        if instance:
            instance = {provider: instance}
        prov_inventory = PROVIDERS[provider](broker_settings=_settings, **instance).get_inventory(
            additional_arg
        )
        curr_inventory = [
            hostname if (hostname := host.get("hostname")) else host.get("name")
            for host in helpers.load_inventory(filter=f'@inv._broker_provider == "{provider}"')
        ]
        helpers.update_inventory(add=prov_inventory, remove=curr_inventory)

    def reconstruct_host(self, host_export_data):
        """Reconstruct a host from export data."""
        logger.debug(f"reconstructing host with export: {host_export_data}")
        provider = PROVIDERS.get(host_export_data.get("_broker_provider"))
        if not provider:
            logger.warning(
                f"No provider found with name {host_export_data.get('_broker_provider')}"
            )
            return
        if provider_instance := host_export_data.get("_broker_provider_instance"):
            host_export_data[provider.__name__] = provider_instance
        provider_inst = provider(broker_settings=self._settings, **host_export_data)
        host = provider_inst.construct_host(
            provider_params=None, host_classes=self.host_classes, **host_export_data
        )
        return host

    def from_inventory(self, filter=None):
        """Reconstruct one or more hosts from the local inventory.

        :param filter: A broker-spec filter string
        """
        inv_hosts = helpers.load_inventory(filter=filter)
        return [self.reconstruct_host(inv_host) for inv_host in inv_hosts]

    @classmethod
    @contextmanager
    def multi_manager(cls, broker_settings=None, **multi_dict):
        """Allow a user to check out multiple hosts at once.

        Given a mapping of names to Broker argument dictionaries:
            create multiple Broker instances, check them out in parallel, yield, then checkin.

        Example:
            with Broker.multi_mode(
                rhel7={
                    "host_class": ContentHost,
                    "workflow": "deploy_base_rhel",
                    "deploy_rhel_version": "7",
                },
                rhel8={
                    "host_class": ContentHost,
                    "workflow": "deploy_base_rhel",
                    "deploy_rhel_version": "8",
                }
            ) as host_dict:
                pass

        All are checked out at the same time. The user is presented with the hosts in
        a dictionary by argument name e.g. host_dict["rhel7"] is a ContentHost object
        """
        # create all the broker instances and perform checkouts in parallel
        _settings = broker_settings or clone_global_settings()
        broker_instances = {
            name: cls(broker_settings=_settings, **kwargs) for name, kwargs in multi_dict.items()
        }
        with ThreadPoolExecutor(max_workers=len(broker_instances)) as workers:
            completed_checkouts = as_completed(
                workers.submit(broker.checkout) for broker in broker_instances.values()
            )
            for completed in completed_checkouts:
                completed.result()
        all_hosts = []
        for broker_inst in broker_instances.values():
            all_hosts.extend(broker_inst._hosts)
        # run setup on all hosts in parallel
        with ThreadPoolExecutor(max_workers=len(all_hosts)) as workers:
            completed_setups = as_completed(workers.submit(host.setup) for host in all_hosts)
            for completed in completed_setups:
                completed.result()
        # yield control to the user
        yield {name: broker._hosts for name, broker in broker_instances.items()}
        # teardown all hosts in parallel
        with ThreadPoolExecutor(max_workers=len(all_hosts)) as workers:
            completed_teardowns = as_completed(workers.submit(host.teardown) for host in all_hosts)
            for completed in completed_teardowns:
                completed.result()
        # checkin all hosts in parallel
        with ThreadPoolExecutor(max_workers=len(broker_instances)) as workers:
            completed_checkins = as_completed(
                workers.submit(broker.checkin) for broker in broker_instances.values()
            )
            for completed in completed_checkins:
                completed.result()

    def __repr__(self):
        """Return a string representation of the Broker instance."""
        inner = ", ".join(
            f"{k}={v}"
            for k, v in self.__dict__.items()
            if not k.startswith("_") and not callable(v)
        )
        return f"{self.__class__.__name__}({inner})"

    def __enter__(self):
        """Checkout hosts and return them to the user."""
        try:
            hosts = self.checkout()
            if not hosts:
                raise Exception("No hosts created during checkout")
            if isinstance(hosts, list):
                [host.setup() for host in hosts]
            else:
                hosts.setup()
            return hosts
        except Exception as err:
            for host in self._hosts:
                _try_teardown(host)
            self.checkin()
            raise err

    def __exit__(self, exc_type, exc_value, exc_traceback):
        """Teardown and checkin hosts."""
        last_exception = None
        for host in self._hosts:
            last_exception = _try_teardown(host)
        self.checkin(in_context=True)
        if last_exception:
            raise last_exception



"""A wrapper around the Beaker CLI."""

import json
from pathlib import Path
import subprocess
import time
from xml.etree import ElementTree as ET

from logzero import logger

from broker import helpers
from broker.exceptions import BeakerBindError


def _elementree_to_dict(etree):
    """Convert an ElementTree object to a dictionary."""
    data = {}
    if etree.attrib:
        data.update(etree.attrib)
    if etree.text:
        data["text"] = etree.text
    for child in etree:
        child_data = _elementree_to_dict(child)
        if (tag := child.tag) in data:
            if not isinstance(data[tag], list):
                data[tag] = [data[tag]]
            data[tag].append(child_data)
        else:
            data[tag] = child_data
    return data


def _curate_job_info(job_info_dict):
    curated_info = {
        "job_id": "id",
        # "reservation_id": "current_reservation/recipe_id",
        "whiteboard": "whiteboard/text",
        "hostname": "recipeSet/recipe/system",
        "distro": "recipeSet/recipe/distro",
    }
    return helpers.dict_from_paths(job_info_dict, curated_info)


class BeakerBind:
    """A bind class providing a basic interface to the Beaker CLI."""

    def __init__(self, hub_url, auth="krbv", **kwargs):
        self.hub_url = hub_url
        self._base_args = ["--insecure", f"--hub={self.hub_url}"]
        if auth == "basic":
            # If we're not using system kerberos auth, add in explicit basic auth
            self.username = kwargs.pop("username", None)
            self.password = kwargs.pop("password", None)
            self._base_args.extend(
                [
                    f"--username {self.username}",
                    f"--password {self.password}",
                ]
            )
        self.__dict__.update(kwargs)

    def _exec_command(self, *cmd_args, **cmd_kwargs):
        """Execute a beaker command and return the result.

        cmd_args: Expanded into feature flags for the beaker command
        cmd_kwargs: Expanded into args and values for the beaker command
        """
        raise_on_error = cmd_kwargs.pop("raise_on_error", True)
        exec_cmd, cmd_args = ["bkr"], list(cmd_args)
        # check through kwargs and if any are True add to cmd_args
        del_keys = []
        for k, v in cmd_kwargs.items():
            if isinstance(v, bool) or v is None:
                del_keys.append(k)
            if v is True:
                cmd_args.append(f"--{k}" if not k.startswith("--") else k)
        for k in del_keys:
            del cmd_kwargs[k]
        exec_cmd.extend(cmd_args)
        exec_cmd.extend(self._base_args)
        exec_cmd.extend([f"--{k.replace('_', '-')}={v}" for k, v in cmd_kwargs.items()])
        logger.debug(f"Executing beaker command: {exec_cmd}")
        proc = subprocess.Popen(
            exec_cmd,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        stdout, stderr = proc.communicate()
        result = helpers.Result(
            stdout=stdout.decode(),
            stderr=stderr.decode(),
            status=proc.returncode,
        )
        if result.status != 0 and raise_on_error:
            raise BeakerBindError(
                f"Beaker command failed:\nCommand={' '.join(exec_cmd)}\nResult={result}",
            )
        logger.debug(f"Beaker command result: {result.stdout}")
        return result

    def job_submit(self, job_xml, wait=False):
        """Submit a job to Beaker and optionally wait for it to complete."""
        # wait behavior seems buggy to me, so best to avoid it
        if not Path(job_xml).exists():
            raise FileNotFoundError(f"Job XML file {job_xml} not found")
        result = self._exec_command("job-submit", job_xml, wait=wait)
        if not wait:
            # get the job id from the output
            # format is "Submitted: ['J:7849837'] where the number is the job id
            for line in result.stdout.splitlines():
                if line.startswith("Submitted:"):
                    return line.split("'")[1].replace("J:", "")

    def job_watch(self, job_id):
        """Watch a job via the job-watch command. This can be buggy."""
        job_id = f"J:{job_id}" if not job_id.startswith("J:") else job_id
        return self._exec_command("job-watch", job_id)

    def job_results(self, job_id, format="beaker-results-xml", pretty=False):
        """Get the results of a job in the specified format."""
        job_id = f"J:{job_id}" if not job_id.startswith("J:") else job_id
        return self._exec_command("job-results", job_id, format=format, prettyxml=pretty)

    def job_clone(self, job_id, wait=False, **kwargs):
        """Clone a job by the specified job id."""
        job_id = f"J:{job_id}" if not job_id.startswith("J:") else job_id
        return self._exec_command("job-clone", job_id, wait=wait, **kwargs)

    def job_list(self, *args, **kwargs):
        """List jobs matching the criteria specified by args and kwargs."""
        return self._exec_command("job-list", *args, **kwargs)

    def job_cancel(self, job_id):
        """Cancel a job by the specified job id."""
        if not job_id.startswith("J:") and not job_id.startswith("RS:"):
            job_id = f"J:{job_id}"
        return self._exec_command("job-cancel", job_id)

    def job_delete(self, job_id):
        """Delete a job by the specified job id."""
        job_id = f"J:{job_id}" if not job_id.startswith("J:") else job_id
        return self._exec_command("job-delete", job_id)

    def system_release(self, system_id):
        """Release a system by the specified system id."""
        return self._exec_command("system-release", system_id)

    def system_list(self, **kwargs):
        """Due to the number of arguments, we will not validate before submitting.

        Accepted arguments are:
        available                       available to be used by this user
        free                            available to this user and not currently being used
        removed                         which have been removed
        mine                            owned by this user
        type=TYPE                       of TYPE
        status=STATUS                   with STATUS
        pool=POOL                       in POOL
        arch=ARCH                       with ARCH
        dev-vendor-id=VENDOR-ID         with a device that has VENDOR-ID
        dev-device-id=DEVICE-ID         with a device that has DEVICE-ID
        dev-sub-vendor-id=SUBVENDOR-ID  with a device that has SUBVENDOR-ID
        dev-sub-device-id=SUBDEVICE-ID  with a device that has SUBDEVICE-ID
        dev-driver=DRIVER               with a device that has DRIVER
        dev-description=DESCRIPTION     with a device that has DESCRIPTION
        xml-filter=XML                  matching the given XML filter
        host-filter=NAME                matching pre-defined host filter
        """
        # convert the flags passed in kwargs to arguments
        args = [
            f"--{key}" for key in ("available", "free", "removed", "mine") if kwargs.pop(key, False)
        ]
        return self._exec_command("system-list", *args, **kwargs)

    def user_systems(self):
        """Return a list of system ids owned by the current user.

        This is used for inventory syncing against Beaker.
        """
        result = self.system_list(mine=True, raise_on_error=False)
        if result.status != 0:
            return []
        else:
            return result.stdout.splitlines()

    def system_details(self, system_id, format="json"):
        """Get details about a system by the specified system id."""
        return self._exec_command("system-details", system_id, format=format)

    def execute_job(self, job, max_wait="24h"):
        """Submit a job, periodically checking the status until it completes.

        return: a dictionary of the results.
        """
        if Path(job).exists():  # job xml path passed in
            job_id = self.job_submit(job, wait=False)
        else:  # using a job id
            job_id = self.job_clone(job)
        logger.info(f"Submitted job: {job_id}")
        _max_wait = time.time() + helpers.translate_timeout(max_wait or "24h")
        while time.time() < _max_wait:
            time.sleep(60)
            result = self.job_results(job_id, pretty=True)
            if 'result="Pass"' in result.stdout:
                return _curate_job_info(_elementree_to_dict(ET.fromstring(result.stdout)))
            elif 'result="Fail"' in result.stdout or "Exception: " in result.stdout:
                raise BeakerBindError(f"Job {job_id} failed:\n{result}")
            elif 'result="Warn"' in result.stdout:
                res_dict = _elementree_to_dict(ET.fromstring(result.stdout))
                raise BeakerBindError(
                    f"Job {job_id} was resulted in a warning. Status: {res_dict['status']}"
                )
        raise BeakerBindError(f"Job {job_id} did not complete within {max_wait}")

    def system_details_curated(self, system_id):
        """Return a curated dictionary of system details."""
        full_details = json.loads(self.system_details(system_id).stdout)
        curated_details = {
            "hostname": full_details["fqdn"],
            "mac_address": full_details["mac_address"],
            "owner": "{display_name} <{email_address}>".format(
                display_name=full_details["owner"]["display_name"],
                email_address=full_details["owner"]["email_address"],
            ),
            "id": full_details["id"],
        }
        if current_res := full_details.get("current_reservation"):
            curated_details.update(
                {
                    "reservation_id": current_res["recipe_id"],
                    "reserved_on": current_res.get("start_time"),
                    "expires_on": current_res.get("finish_time"),
                    "reserved_for": "{display_name} <{email_address}>".format(
                        display_name=current_res["user"]["display_name"],
                        email_address=current_res["user"]["email_address"],
                    ),
                }
            )
        return curated_details

    def jobid_from_system(self, system_hostname):
        """Return the job id for the current reservation on the system."""
        for job_id in json.loads(self.job_list(mine=True).stdout):
            job_result = self.job_results(job_id, pretty=True)
            job_detail = _curate_job_info(_elementree_to_dict(ET.fromstring(job_result.stdout)))
            if job_detail["hostname"] == system_hostname:
                return job_id



"""Module handling internal and dependency logging."""

import copy
from enum import IntEnum
import logging

import logzero

from broker.settings import BROKER_DIRECTORY


class LOG_LEVEL(IntEnum):
    """Bare class for log levels. Trace is added for custom logging."""

    TRACE = 5
    DEBUG = logging.DEBUG
    INFO = logging.INFO
    WARNING = logging.WARNING
    ERROR = logging.ERROR


DEFAULT_FILE_LEVEL = LOG_LEVEL.DEBUG  # Default file logging level if not overridden


class RedactingFilter(logging.Filter):
    """Custom logging.Filter to redact secrets from the Dynaconf config."""

    def __init__(self, sensitive):
        super().__init__()
        self._sensitive = sensitive

    def filter(self, record):
        """Filter the record and redact the sensitive keys."""
        if isinstance(record.args, dict):
            record.args = self.redact_dynaconf(record.args)
        else:
            record.args = tuple(self.redact_dynaconf(arg) for arg in record.args)
        return True

    def redact_dynaconf(self, data):
        """Go over the data and redact all values of keys that match the sensitive ones."""
        if isinstance(data, list | tuple):
            data_copy = [self.redact_dynaconf(item) for item in data]
        elif isinstance(data, dict):
            data_copy = copy.deepcopy(data)
            for k, v in data_copy.items():
                if isinstance(v, dict | list):
                    data_copy[k] = self.redact_dynaconf(v)
                elif k in self._sensitive and v:
                    data_copy[k] = "******"
        else:
            data_copy = data
        return data_copy


_sensitive = ["password", "pword", "token", "host_password"]
logging.addLevelName("TRACE", LOG_LEVEL.TRACE)
logzero.DEFAULT_COLORS[LOG_LEVEL.TRACE.value] = logzero.colors.Fore.MAGENTA


def try_disable_urllib3_warnings():
    """Attempt to disable urllib3 InsecureRequestWarning if urllib3 is available."""
    try:
        import urllib3
    except ImportError:
        logzero.logger.debug("urllib3 not installed, skipping urllib3 warning patch.")
        return
    urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)


def try_patch_awx_for_verbosity():
    """Patch the awxkit API to enable trace-level logging of API calls to Ansible provider."""
    try:
        from awxkit import api
    except ImportError:
        logzero.logger.debug("awxkit not installed, skipping awxkit logging patch.")
        return
    awx_log = api.client.log
    awx_log.parent = logzero.logger

    def patch(cls, name):
        func = getattr(cls, name)

        def the_patch(self, *args, **kwargs):
            awx_log.log(LOG_LEVEL.TRACE.value, f"Calling {self=} {func=}(*{args=}, **{kwargs=}")
            retval = func(self, *args, **kwargs)
            awx_log.log(
                LOG_LEVEL.TRACE.value,
                f"Finished {self=} {func=}(*{args=}, **{kwargs=}) {retval=}",
            )
            return retval

        setattr(cls, name, the_patch)

    for method in "delete get head options patch post put".split():
        patch(api.Connection, method)


def resolve_log_level(level):
    """Resolve the log level from a string."""
    try:
        log_level = LOG_LEVEL[level.upper()]
    except KeyError:
        log_level = LOG_LEVEL.INFO
    return log_level


def formatter_factory(log_level, color=True):
    """Create a logzero formatter based on the log level."""
    log_fmt = "%(color)s[%(levelname)s %(asctime)s]%(end_color)s %(message)s"
    debug_fmt = (
        "%(color)s[%(levelname)1.1s %(asctime)s %(module)s:%(lineno)d]%(end_color)s %(message)s"
    )
    formatter = logzero.LogFormatter(
        fmt=debug_fmt if log_level <= LOG_LEVEL.DEBUG else log_fmt, color=color
    )
    return formatter


def set_log_level(level):
    """Set the log level for logzero."""
    log_level = LOG_LEVEL.INFO if level == "silent" else resolve_log_level(level)
    logzero.formatter(formatter=formatter_factory(log_level))
    logzero.loglevel(level=log_level)


def set_file_logging(level, path="logs/broker.log"):
    """Set the file logging for logzero."""
    silent = False
    if level == "silent":
        silent = True
        log_level = LOG_LEVEL.INFO
    else:
        # Allow override of file logging level with if the new level is lower than the default.
        # Otherwise, use the default value.
        new_log_level = resolve_log_level(level)
        log_level = new_log_level if new_log_level < DEFAULT_FILE_LEVEL else DEFAULT_FILE_LEVEL

    path = BROKER_DIRECTORY.joinpath(path)
    path.parent.mkdir(parents=True, exist_ok=True)
    logzero.logfile(
        path,
        loglevel=log_level.value,
        maxBytes=1e9,
        backupCount=3,
        formatter=formatter_factory(log_level, color=False),
        disableStderrLogger=silent,
    )


def setup_logzero(
    level="info",
    formatter=None,
    file_level="debug",
    name=None,
    path="logs/broker.log",
):
    """Call logzero setup with the given settings."""
    level = level or "info"
    file_level = file_level or "debug"
    path = path or "logs/broker.log"
    set_log_level(level)
    set_file_logging(file_level, path)
    if formatter:
        logzero.formatter(formatter)
    logzero.logger.name = name or "broker"
    logzero.logger.addFilter(RedactingFilter(_sensitive))


try_disable_urllib3_warnings()
try_patch_awx_for_verbosity()
setup_logzero()



"""Beaker provider implementation."""

import inspect

import click
from dynaconf import Validator
from logzero import logger

from broker import helpers
from broker.binds.beaker import BeakerBind
from broker.exceptions import BrokerError, ProviderError
from broker.hosts import Host
from broker.providers import Provider


class Beaker(Provider):
    """Beaker provider class providing a Broker interface around the Beaker bind."""

    _validators = [
        Validator("beaker.hub_url", must_exist=True),
        Validator("beaker.max_job_wait", default="24h"),
    ]
    _checkout_options = [
        click.option(
            "--job-xml",
            type=click.Path(exists=True, dir_okay=False),
            help="Path to the job XML file to submit",
        ),
        click.option(
            "--job-id",
            type=str,
            help="Beaker job ID to clone",
        ),
    ]
    _execute_options = [
        click.option(
            "--job-xml",
            type=str,
            help="Path to the job XML file to submit",
        ),
    ]
    _extend_options = [
        click.option(
            "--extend-duration",
            type=click.IntRange(1, 99),
            help="Number of hours to extend the job. Must be between 1 and 99",
        )
    ]

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.hub_url = self._settings.beaker.hub_url
        self.runtime = kwargs.pop("bind", BeakerBind)(self.hub_url, **kwargs)

    def _host_release(self):
        caller_host = inspect.stack()[1][0].f_locals["host"]
        if not (job_id := getattr(caller_host, "job_id", None)):
            job_id = self.runtime.jobid_from_system(caller_host.hostname)
        return self.release(caller_host.hostname, job_id)

    def _set_attributes(self, host_inst, broker_args=None, misc_attrs=None):
        host_inst.__dict__.update(
            {
                "_prov_inst": self,
                "_broker_provider": "Beaker",
                "_broker_provider_instance": self.instance,
                "_broker_args": broker_args,
                "release": self._host_release,
            }
        )
        if isinstance(misc_attrs, dict):
            host_inst._attrs = misc_attrs

    def _compile_host_info(self, host, broker_info=True):
        """Compiles host information into a dictionary suitable for use in the inventory.

        :param host (beaker.host.Host): The host to compile information for.

        :return: A dictionary containing the compiled host information.
        """
        curated_host_info = self.runtime.system_details_curated(host)
        if broker_info:
            curated_host_info.update(
                {
                    "_broker_provider": "Beaker",
                    "_broker_provider_instance": self.instance,
                    "_broker_args": getattr(host, "_broker_args", {}),
                }
            )
        if not curated_host_info.get("job_id"):
            curated_host_info["job_id"] = self.runtime.jobid_from_system(
                curated_host_info["hostname"]
            )
        return curated_host_info

    def construct_host(self, provider_params, host_classes, **kwargs):
        """Construct a broker host from a beaker system information.

        :param provider_params: a beaker system information dictionary

        :param host_classes: host object

        :return: constructed broker host object
        """
        logger.debug(f"constructing with {provider_params=}\n{host_classes=}\n{kwargs=}")
        if not provider_params:
            host_inst = host_classes[kwargs.get("type", "host")](**kwargs)
            # cont_inst = self._cont_inst_by_name(host_inst.name)
            self._set_attributes(host_inst, broker_args=kwargs)
        else:
            host_info = self._compile_host_info(provider_params["hostname"], broker_info=False)
            host_inst = host_classes[kwargs.get("type", "host")](
                broker_settings=self._settings, **provider_params
            )
            self._set_attributes(host_inst, broker_args=kwargs, misc_attrs=host_info)
        return host_inst

    @Provider.register_action("job_xml", "job_id")
    def submit_job(self, max_wait=None, **kwargs):
        """Submit a job to Beaker and wait for it to complete."""
        job = kwargs.get("job_xml") or kwargs.get("job_id")
        max_wait = max_wait or self._settings.beaker.get("max_job_wait")
        result = self.runtime.execute_job(job, max_wait)
        logger.debug(f"Job completed with results: {result}")
        return result

    def provider_help(self, jobs=False, job=None, **kwargs):
        """Print useful information from the Beaker provider."""
        results_limit = kwargs.get("results_limit", self._settings.container.results_limit)
        if job:
            if not job.startswith("J:"):
                job = f"J:{job}"
            logger.info(self.runtime.job_clone(job, prettyxml=True, dryrun=True).stdout)
        elif jobs:
            result = self.runtime.job_list(**kwargs).stdout.splitlines()
            if res_filter := kwargs.get("results_filter"):
                result = helpers.eval_filter(result, res_filter, "res")
            result = "\n".join(result[:results_limit])
            logger.info(f"Available jobs:\n{result}")

    def release(self, host_name, job_id):
        """Release a hosts reserved from Beaker by cancelling the job."""
        return self.runtime.job_cancel(job_id)
        # return self.runtime.system_release(host_name)

    def extend(self, host_name, extend_duration=99):
        """Extend the duration of a Beaker reservation."""
        try:
            Host(hostname=host_name).execute(f"/usr/bin/extendtesttime.sh {extend_duration}")
        except BrokerError as err:
            raise ProviderError(
                f"Failed to extend host {host_name}: {err}\n"
                f"Try running: root@{host_name} /usr/bin/extendtesttime.sh {extend_duration}"
            ) from err

    def get_inventory(self, *args):
        """Get a list of hosts and their information from Beaker."""
        hosts = self.runtime.user_systems()
        with click.progressbar(hosts, label="Compiling host information") as hosts_bar:
            compiled_host_info = [self._compile_host_info(host) for host in hosts_bar]
        return compiled_host_info



"""OpenStack provider implementation."""

from contextlib import contextmanager
import inspect
import logging
from uuid import uuid4

import click
from dynaconf import Validator
from logzero import logger

from broker import exceptions, helpers
from broker.providers import Provider

# Deferred imports - will be imported when first OpenStack instance is created
openstack = None
OpenStackCloudException = None
ResourceNotFound = None


@contextmanager
def _suppress_retry_warnings():
    """Suppress retry warnings during server creation."""
    original_level = logger.level
    logger.setLevel(logging.ERROR)
    try:
        yield
    finally:
        logger.setLevel(original_level)


class OpenStack(Provider):
    """OpenStack provider class providing a Broker interface around OpenStack SDK."""

    _validators = [
        Validator("OPENSTACK.server_timeout", default=600),
        Validator("OPENSTACK.user_domain_name", default="Default"),
        Validator("OPENSTACK.project_domain_name", default="Default"),
        Validator("OPENSTACK.identity_api_version", default="3"),
    ]
    _checkout_options = [
        click.option("--ostack-image", help="Image name or UUID for the VM"),
        click.option("--ostack-flavor", help="Flavor for the VM"),
        click.option("--ostack-network", help="Network UUID for the VM"),
        click.option("--ostack-key-name", help="SSH key pair name"),
        click.option("--ostack-template", help="Template name from configuration"),
    ]

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        # Import openstack dependencies on first use
        global openstack, OpenStackCloudException, ResourceNotFound  # noqa: PLW0603
        if openstack is None:
            try:
                import openstack
                from openstack.exceptions import OpenStackCloudException, ResourceNotFound
            except ImportError as e:
                raise ImportError(
                    "openstacksdk is required for the OpenStack provider. "
                    "Please install it with 'pip install openstacksdk' or 'pip install broker[openstack]'"
                ) from e
        self._connection = None

    @property
    def connection(self):
        """Get OpenStack connection."""
        if self._connection is None:
            self._connection = self._get_connection()
        return self._connection

    @connection.setter
    def connection(self, value):
        """Set OpenStack connection."""
        self._connection = value

    def _get_connection_with_cloud(self, cloud_name):
        """Connect using cloud configuration."""
        logger.debug(f"Connecting to OpenStack cloud: {cloud_name}")
        return openstack.connect(cloud=cloud_name)

    def _get_connection_with_app_credentials(self, auth_url):
        """Connect using application credentials."""
        app_cred_id = getattr(self._settings.OPENSTACK, "application_credential_id", None)
        app_cred_secret = getattr(self._settings.OPENSTACK, "application_credential_secret", None)

        if not (app_cred_id and app_cred_secret):
            return None

        auth_config = {
            "auth_url": auth_url,
            "application_credential_id": app_cred_id,
            "application_credential_secret": app_cred_secret,
            "identity_api_version": self._settings.OPENSTACK.identity_api_version,
        }

        for param in ["region_name", "interface"]:
            if hasattr(self._settings.OPENSTACK, param):
                auth_config[param] = getattr(self._settings.OPENSTACK, param)

        logger.debug(f"Connecting with app credentials: {auth_url}")
        return openstack.connect(**auth_config)

    def _get_connection_with_password(self, auth_url):
        """Connect with username/password."""
        username = getattr(self._settings.OPENSTACK, "username", None)
        password = getattr(self._settings.OPENSTACK, "password", None)
        project_name = getattr(self._settings.OPENSTACK, "project_name", None)

        if not all([username, password, project_name]):
            return None

        auth_config = {
            "auth_url": auth_url,
            "username": username,
            "password": password,
            "project_name": project_name,
            "user_domain_name": self._settings.OPENSTACK.user_domain_name,
            "project_domain_name": self._settings.OPENSTACK.project_domain_name,
            "identity_api_version": self._settings.OPENSTACK.identity_api_version,
        }

        for param in ["region_name", "interface"]:
            if hasattr(self._settings.OPENSTACK, param):
                auth_config[param] = getattr(self._settings.OPENSTACK, param)

        logger.debug(f"Connecting with username/password: {auth_url}")
        return openstack.connect(**auth_config)

    def _get_connection(self):
        """Establish OpenStack connection."""
        try:
            cloud_name = getattr(self._settings.OPENSTACK, "cloud", None)
            if cloud_name:
                return self._get_connection_with_cloud(cloud_name)

            auth_url = getattr(self._settings.OPENSTACK, "auth_url", None)
            if not auth_url:
                raise exceptions.ConfigurationError(
                    "OpenStack config incomplete. Provide 'cloud' or 'auth_url'"
                )

            connection = self._get_connection_with_app_credentials(auth_url)
            if connection:
                return connection

            connection = self._get_connection_with_password(auth_url)
            if connection:
                return connection

            raise exceptions.ConfigurationError(
                "OpenStack config incomplete. Provide one of:\n"
                "1. cloud (clouds.yaml)\n"
                "2. auth_url + app credentials\n"
                "3. auth_url + username/password"
            )
        except (OpenStackCloudException, exceptions.ConfigurationError) as e:
            logger.error(f"Failed to connect to OpenStack: {e!s}")
            raise exceptions.ProviderError("OpenStack", f"Failed to connect: {e!s}") from e

    def _host_release(self):
        """Release method for host objects."""
        caller_host = inspect.stack()[1][0].f_locals["host"]
        return self.release(caller_host)

    def _set_attributes(self, host_inst, broker_args=None):
        """Set host attributes."""
        host_inst.__dict__.update(
            {
                "_prov_inst": self,
                "_broker_provider": "OpenStack",
                "_broker_provider_instance": self.instance,
                "_broker_args": broker_args,
                "release": self._host_release,
            }
        )

        if broker_args and "instance_id" in broker_args and not hasattr(host_inst, "instance_id"):
            host_inst.instance_id = broker_args["instance_id"]

    def provider_help(self, images=False, flavors=False, networks=False, templates=False, **kwargs):
        """Display OpenStack provider information."""
        if images:
            logger.info("Available images:")
            for image in self.connection.image.images():
                if image.status == "active":
                    logger.info(f"  - {image.name} ({image.id})")
        elif flavors:
            logger.info("Available flavors:")
            for flavor in self.connection.compute.flavors():
                logger.info(f"  - {flavor.name} ({flavor.id})")
        elif networks:
            logger.info("Available networks:")
            for network in self.connection.network.networks():
                logger.info(f"  - {network.name} ({network.id})")
        elif templates:
            templates = self._settings.OPENSTACK.get("templates", {})
            if templates:
                logger.info("Available templates:")
                for template_name, template_config in templates.items():
                    logger.info(f"  - {template_name}: {template_config}")
            else:
                logger.info("No templates configured")
        else:
            logger.info("OpenStack provider configured and connected")

    def construct_host(self, provider_params, host_classes, **kwargs):
        """Construct a host object from the provider_params and kwargs."""
        logger.debug(f"constructing with {provider_params=}\n{host_classes=}\n{kwargs=}")

        if provider_params:
            host_inst = host_classes[kwargs.get("type", "host")](**provider_params)
            if "instance_id" in provider_params:
                host_inst.instance_id = provider_params["instance_id"]
        else:
            host_inst = host_classes[kwargs.get("type", "host")](**kwargs)
            if "instance_id" in kwargs:
                host_inst.instance_id = kwargs["instance_id"]

        self._set_attributes(host_inst, broker_args=kwargs)

        if not hasattr(host_inst, "instance_id") or not host_inst.instance_id:
            logger.debug(
                f"Host object {host_inst} does not have instance_id set. Available attributes: {dir(host_inst)}"
            )

        return host_inst

    def _get_template_params(self, template_name, kwargs):
        """Get parameters from template configuration."""
        templates = self._settings.OPENSTACK.get("templates", {})
        if template_name not in templates:
            raise exceptions.UserError(
                f"Template '{template_name}' not found in OpenStack configuration"
            )

        template = templates[template_name]
        params = template.copy()
        param_mapping = {
            "ostack_image": "image",
            "ostack_flavor": "flavor",
            "ostack_network": "network",
            "ostack_key_name": "key_name",
        }
        for cli_key, param_key in param_mapping.items():
            if cli_key in kwargs:
                params[param_key] = kwargs[cli_key]
        return params

    def _get_direct_params(self, kwargs):
        """Get direct parameters with defaults."""
        params = {}

        param_mapping = {
            "ostack_image": "image",
            "ostack_flavor": "flavor",
            "ostack_network": "network",
            "ostack_key_name": "key_name",
            "name": "name",
        }

        for cli_key, param_key in param_mapping.items():
            if cli_key in kwargs:
                params[param_key] = kwargs[cli_key]

        default_mapping = {
            "image": "default_image",
            "flavor": "default_flavor",
            "network": "default_network",
            "key_name": "default_key_name",
        }

        for param, default_key in default_mapping.items():
            if param not in params:
                params[param] = getattr(self._settings.OPENSTACK, default_key, None)

        return params

    @Provider.register_action("ostack_template", "ostack_image")
    def checkout(self, **kwargs):
        """Create an OpenStack instance."""
        logger.debug(f"Checkout kwargs: {kwargs}")

        template_name = kwargs.get("ostack_template")

        if template_name:
            params = self._get_template_params(template_name, kwargs)
        else:
            params = self._get_direct_params(kwargs)

        required_params = ["image", "flavor", "network"]
        missing_params = [param for param in required_params if not params.get(param)]

        if missing_params:
            missing_str = ", ".join(missing_params)
            raise exceptions.UserError(
                f"Missing required OpenStack parameters: {missing_str}. "
                f"Provide them directly or via template/defaults in configuration."
            )

        logger.info(
            f"Creating OpenStack server with image: {params['image']}, "
            f"flavor: {params['flavor']}, network: {params['network']}"
        )

        try:
            image_id = self._resolve_image(params["image"])
            flavor_id = self._resolve_flavor(params["flavor"])
            network_id = self._resolve_network(params["network"])

            server_params = {
                "name": params.get("name", f"broker-{str(uuid4()).split('-')[0]}"),
                "image_id": image_id,
                "flavor_id": flavor_id,
                "networks": [{"uuid": network_id}],
                "wait": False,
            }

            if params.get("key_name"):
                server_params["key_name"] = params["key_name"]

            logger.debug(f"Server creation parameters: {server_params}")

            server = self.connection.compute.create_server(**server_params)
            logger.info(f"Server creation initiated: {server.id}")

            server_id = server.id
            timeout = self._settings.OPENSTACK.server_timeout

            def check_server():
                s = self.connection.compute.get_server(server_id)
                logger.debug(f"Server {server_id} status: {s.status}")
                if s.status == "ACTIVE":
                    logger.info(f"Server {server_id} is now active")
                    return s
                elif s.status == "ERROR":
                    raise exceptions.ProviderError(
                        "OpenStack", f"Server {server_id} failed to build"
                    )
                else:
                    logger.info(f"Server {server_id} building... (status: {s.status})")
                    raise Exception("Server not ready yet")

            with _suppress_retry_warnings():
                server = helpers.simple_retry(
                    check_server,
                    max_timeout=timeout,
                    terminal_exceptions=(exceptions.ProviderError,),
                )

            if not server:
                raise exceptions.ProviderError(
                    "OpenStack", f"Failed to retrieve server {server_id} after creation"
                )

            ip_address = self._get_ip_address(server)
            logger.info(f"Server {server.id} assigned IP: {ip_address}")

            provider_params = {
                "hostname": ip_address,
                "instance_id": server.id,
                "name": server.name,
            }

            return provider_params

        except (OpenStackCloudException, ResourceNotFound) as e:
            logger.error(f"Failed to create OpenStack server: {e!s}")
            raise exceptions.ProviderError("OpenStack", f"Failed to create server: {e!s}") from e

    def _get_ip_address(self, server):
        """Extract IP address from server object."""
        for addresses in server.addresses.values():
            for address in addresses:
                if address.get("OS-EXT-IPS:type") == "floating":
                    return address["addr"]

        for addresses in server.addresses.values():
            for address in addresses:
                return address["addr"]

        return server.name

    def _resolve_image(self, image_name_or_id):
        """Resolve image name to UUID."""
        try:
            try:
                image = self.connection.image.get_image(image_name_or_id)
                logger.debug(f"Image resolved by ID: {image_name_or_id}")
                return image.id
            except (ResourceNotFound, OpenStackCloudException):
                logger.debug(f"Image '{image_name_or_id}' not found by ID, trying name lookup")

            images = list(self.connection.image.images(name=image_name_or_id))
            if images:
                image_id = images[0].id
                logger.debug(f"Image '{image_name_or_id}' resolved to ID: {image_id}")
                return image_id

            all_images = list(self.connection.image.images())
            matching_images = [img for img in all_images if image_name_or_id in img.name]

            if matching_images:
                image_id = matching_images[0].id
                logger.debug(
                    f"Image '{image_name_or_id}' partially matched to '{matching_images[0].name}' (ID: {image_id})"
                )
                return image_id

            raise exceptions.UserError(f"Image '{image_name_or_id}' not found in OpenStack")

        except OpenStackCloudException as e:
            logger.error(f"Failed to resolve image '{image_name_or_id}': {e!s}")
            raise exceptions.ProviderError("OpenStack", f"Failed to resolve image: {e!s}") from e

    def _resolve_flavor(self, flavor_name_or_id):
        """Resolve flavor name to UUID."""
        try:
            try:
                flavor = self.connection.compute.get_flavor(flavor_name_or_id)
                logger.debug(f"Flavor resolved by ID: {flavor_name_or_id}")
                return flavor.id
            except (ResourceNotFound, OpenStackCloudException):
                logger.debug(f"Flavor '{flavor_name_or_id}' not found by ID, trying name lookup")

            flavors = list(self.connection.compute.flavors(name=flavor_name_or_id))
            if flavors:
                flavor_id = flavors[0].id
                logger.debug(f"Flavor '{flavor_name_or_id}' resolved to ID: {flavor_id}")
                return flavor_id

            raise exceptions.UserError(f"Flavor '{flavor_name_or_id}' not found in OpenStack")

        except OpenStackCloudException as e:
            logger.error(f"Failed to resolve flavor '{flavor_name_or_id}': {e!s}")
            raise exceptions.ProviderError("OpenStack", f"Failed to resolve flavor: {e!s}") from e

    def _resolve_network(self, network_name_or_id):
        """Resolve network name to UUID."""
        try:
            try:
                network = self.connection.network.get_network(network_name_or_id)
                logger.debug(f"Network resolved by ID: {network_name_or_id}")
                return network.id
            except (ResourceNotFound, OpenStackCloudException):
                logger.debug(f"Network '{network_name_or_id}' not found by ID, trying name lookup")

            networks = list(self.connection.network.networks(name=network_name_or_id))
            if networks:
                network_id = networks[0].id
                logger.debug(f"Network '{network_name_or_id}' resolved to ID: {network_id}")
                return network_id

            raise exceptions.UserError(f"Network '{network_name_or_id}' not found in OpenStack")

        except OpenStackCloudException as e:
            logger.error(f"Failed to resolve network '{network_name_or_id}': {e!s}")
            raise exceptions.ProviderError("OpenStack", f"Failed to resolve network: {e!s}") from e

    def _find_server_by_name_or_ip(self, hostname, name):
        """Find OpenStack server by name or IP address."""
        try:
            servers = list(self.connection.compute.servers(name=name))
            if servers:
                logger.info(f"Found server by name '{name}': {servers[0].id}")
                return servers[0].id

            all_servers = list(self.connection.compute.servers())
            for server in all_servers:
                for addresses in server.addresses.values():
                    for address in addresses:
                        if address.get("addr") == hostname:
                            logger.info(f"Found server by IP '{hostname}': {server.id}")
                            return server.id
            return None
        except OpenStackCloudException as e:
            logger.warning(f"Failed to search for server: {e}")
            return None

    def release(self, host_obj):
        """Release (delete) an OpenStack server."""
        instance_id = None

        if hasattr(host_obj, "instance_id") and host_obj.instance_id:
            instance_id = host_obj.instance_id
        elif hasattr(host_obj, "server_id") and host_obj.server_id:
            instance_id = host_obj.server_id
        elif hasattr(host_obj, "id") and host_obj.id:
            instance_id = host_obj.id
        elif hasattr(host_obj, "_broker_args") and host_obj._broker_args:
            logger.debug(f"Checking _broker_args: {host_obj._broker_args}")
            if isinstance(host_obj._broker_args, dict):
                instance_id = host_obj._broker_args.get("instance_id")

        if not instance_id:
            logger.info(f"Attempting to find OpenStack server by hostname: {host_obj.hostname}")
            instance_id = self._find_server_by_name_or_ip(host_obj.hostname, host_obj.name)

        if not instance_id:
            logger.error(
                f"Cannot find instance ID for host {host_obj}. Available attributes: {dir(host_obj)}"
            )
            raise exceptions.ProviderError(
                "OpenStack", f"Cannot find instance ID for host {host_obj}"
            )

        try:
            logger.info(f"Releasing OpenStack server: {instance_id}")
            self.connection.compute.delete_server(instance_id)
            logger.info(f"Server {instance_id} deleted successfully")
            return True
        except OpenStackCloudException as e:
            logger.error(f"Failed to delete server {instance_id}: {e!s}")
            raise exceptions.ProviderError("OpenStack", f"Failed to delete server: {e!s}") from e

    def extend(self, host_obj):
        """Extend lease time for an OpenStack server (if supported)."""
        raise NotImplementedError(
            "Extend operation is not implemented for OpenStack provider. "
            "OpenStack doesn't have a built-in lease extension concept."
        )

    def get_inventory(self, **kwargs):
        """Get inventory of OpenStack servers."""
        try:
            servers = list(self.connection.compute.servers())
            inventory = []

            for server in servers:
                ip_address = self._get_ip_address(server)
                inventory.append(
                    {
                        "hostname": ip_address,
                        "instance_id": server.id,
                        "name": server.name,
                        "status": server.status,
                        "provider": "OpenStack",
                    }
                )

            return inventory
        except OpenStackCloudException as e:
            logger.error(f"Failed to get OpenStack inventory: {e!s}")
            raise exceptions.ProviderError("OpenStack", f"Failed to get inventory: {e!s}") from e



"""Module providing classes to establish ssh or ssh-like connections to hosts.

Classes:
    Session - Wrapper around ansible-pylibssh auth/connection system.
    InteractiveShell - Wrapper around ansible-pylibssh non-blocking channel system.

Note: You typically want to use a Host object instance to create sessions,
      not these classes directly.
"""

from contextlib import contextmanager
from pathlib import Path
from tempfile import NamedTemporaryFile

from pylibsshext.session import Session as _Session

from broker import exceptions, helpers
from broker.binds.utils import _create_connect_socket


class Session:
    """Wrapper around ansible-pylibssh's auth/connection system."""

    def __init__(self, **kwargs):
        """Initialize a Session object.

        kwargs:
            hostname (str): The hostname or IP address of the remote host. Defaults to 'localhost'.
            username (str): The username to authenticate with. Defaults to 'root'.
            timeout (float): The timeout for the connection in seconds. Defaults to 60.
            port (int): The port number to connect to. Defaults to 22.
            key_filename (str): The path to the private key file to use for authentication.
            password (str): The password to use for authentication.
            ipv6 (bool): Whether or not to use IPv6. Defaults to False.
            ipv4_fallback (bool): Whether or not to fallback to IPv4 if IPv6 fails. Defaults to True.

        Raises:
            AuthException: If no password or key file is provided.
            ConnectionError: If the connection fails.
            FileNotFoundError: If the key file is not found.
        """
        host = kwargs.get("hostname", "localhost")
        user = kwargs.get("username", "root")
        port = kwargs.get("port", 22)
        key_filename = kwargs.get("key_filename")
        password = kwargs.get("password")
        timeout = kwargs.get("timeout", 60)

        # Create the socket
        self.sock, self.is_ipv6 = _create_connect_socket(
            host,
            port,
            timeout,
            ipv6=kwargs.get("ipv6", False),
            ipv4_fallback=kwargs.get("ipv4_fallback", True),
        )

        self.session = _Session()
        try:
            if key_filename:
                auth_type = "Key"
                key_path = Path(key_filename)
                if not key_path.exists():
                    raise FileNotFoundError(f"Key not found in '{key_filename}'")
                self.session.connect(
                    fd=self.sock.fileno(),
                    host=host,
                    host_key_checking=False,
                    port=port,
                    private_key=key_path.read_bytes(),
                    timeout=timeout,
                    user=user,
                )
            elif password:
                auth_type = "Password"
                self.session.connect(
                    fd=self.sock.fileno(),
                    host=host,
                    host_key_checking=False,
                    password=password,
                    port=port,
                    timeout=timeout,
                    user=user,
                )
            elif user:
                auth_type = "Session"
                raise exceptions.NotImplementedError("Session-based auth for ansible-pylibssh")
            else:
                raise exceptions.AuthenticationError("No password or key file provided.")
        except Exception as err:
            raise exceptions.AuthenticationError(
                f"{auth_type}-based authentication failed."
            ) from err

    @staticmethod
    def _set_destination(source, destination):
        dest = destination or source
        if dest.endswith("/"):
            dest = dest + Path(source).name
        return dest

    def disconnect(self):
        """Disconnect session."""

    def remote_copy(self, source, dest_host, dest_path=None, ensure_dir=True):
        """Copy a file from this host to another."""
        dest_path = dest_path or source
        if ensure_dir:
            dest_host.session.run(f"mkdir -p {Path(dest_path).absolute().parent}")

        # TODO read/write without local dest_path intermediate
        sftp_down = self.session.sftp()
        sftp_up = dest_host.session.session.sftp()
        try:
            with NamedTemporaryFile() as tmp:
                sftp_down.get(source, tmp.file.name)
                sftp_up.put(tmp.file.name, dest_path)
        finally:
            sftp_down.close()
            sftp_up.close()

    def run(self, command, timeout=0):
        """Run a command on the host and return the results."""
        channel = self.session.new_channel()
        try:
            res = channel.exec_command(command)
            return helpers.Result(
                status=res.returncode,
                stdout=res.stdout.decode("utf-8"),
                stderr=res.stderr.decode("utf-8"),
            )
        finally:
            channel.close()

    def scp_write(self, source, destination=None, ensure_dir=True):
        """SCP write a local file to a remote destination."""
        destination = self._set_destination(source, destination)
        if ensure_dir:
            self.run(f"mkdir -p {Path(destination).absolute().parent}")

        scp = self.session.scp()
        scp.put(destination, source)

    def sftp_read(self, source, destination=None, return_data=False):
        """Read a remote file into a local destination or return a bytes object if return_data is True."""
        # TODO read contents directly into bytes object if return_data is True
        destination = self._set_destination(source, destination)
        # Create the destination path if it doesn't exist
        destination = Path(destination)
        destination.parent.mkdir(parents=True, exist_ok=True)

        # Initiate the sftp session, read data, write it to a local destination
        sftp = self.session.sftp()
        try:
            sftp.get(source, destination)
            if return_data:
                return destination.read_bytes()
        finally:
            if return_data:
                destination.unlink()
            sftp.close()

    def sftp_write(self, source, destination=None, ensure_dir=True):
        """Sftp write a local file to a remote destination."""
        destination = self._set_destination(source, destination)
        if ensure_dir:
            self.run(f"mkdir -p {Path(destination).absolute().parent}")

        sftp = self.session.sftp()
        try:
            sftp.put(source, destination)
        finally:
            sftp.close()

    def shell(self, pty=False):
        """Create and return an interactive shell instance."""
        return InteractiveShell(self.session, pty=pty)

    @contextmanager
    def tail_file(self, filename):
        """Tail a file on the remote host."""
        # TODO re-factor to use SFTP instead
        initial_size = int(self.run(f"stat -c %s {filename}").stdout.strip())
        yield (tailer := FileTailer(session=self.session, filename=filename))
        tailer.contents = self.run(f"tail -c +{initial_size} {filename}").stdout


class FileTailer:
    """FileTailer class."""

    def __init__(self, **kwargs):
        self.session = kwargs.get("session")
        self.filename = kwargs.get("filename")


class InteractiveShell:
    """A helper class that provides an interactive shell interface.

    Preferred use of this class is via its context manager

    with InteractiveShell(my_session) as shell:
        shell.send("some-command --argument")
        shell.send("another-command")
        time.sleep(5)  # give time for things to complete
    assert "expected text" in shell.result.stdout

    """

    def __init__(self, session, pty=False):
        # FIXME: invoke_shell() always requests pty
        # self._channel = session.invoke_shell(pty=pty)
        if pty:
            self._channel = session.invoke_shell()
        else:
            raise exceptions.NotImplementedError("Interactive shell with pty=False")

    def __enter__(self):
        """Return the shell object."""
        return self

    def __exit__(self, *exc_args):
        """Close the channel and read stdout/stderr and status."""
        self.send("exit")  # ensure shell has exited
        self._channel.send_eof()

        stdout = self._channel.read_bulk_response(timeout=0.5)
        stderr = self._channel.read_bulk_response(stderr=1)
        status = self._channel.get_channel_exit_status()

        self._channel.close()

        self.result = helpers.Result(
            status=status,
            stdout=stdout.decode("utf-8"),
            stderr=stderr.decode("utf-8"),
        )

    def __getattribute__(self, name):
        """Expose non-duplicate attributes from the Channel instance."""
        try:
            return object.__getattribute__(self, name)
        except AttributeError:
            return getattr(self._channel, name)

    def send(self, cmd):
        """Send a command to the channel, ensuring a newline character."""
        if not cmd.endswith("\n"):
            cmd += "\n"
        self._channel.write(cmd.encode("utf-8"))

    def stdout(self):
        """Read the contents of a channel's stdout."""
        # FIXME handle read on open channel
        res = self._channel.read_bulk_response()
        return res.stdout.decode("utf-8")



"""Module providing classes to establish ssh or ssh-like connections to hosts.

Classes:
    Session - Wrapper around ssh2-python's auth/connection system.
    InteractiveShell - Wrapper around ssh2-python's non-blocking channel system.
    ContainerSession - Wrapper around docker-py's exec system.

Note: You typically want to use a Host object instance to create sessions,
      not these classes directly.
"""

from contextlib import contextmanager
from importlib.metadata import entry_points
from pathlib import Path
import tempfile

from logzero import logger

from broker import helpers
from broker.exceptions import NotImplementedError
from broker.settings import clone_global_settings

SSH_NOT_INSTALLED_MSG = (
    "{backend} is not installed.\n"
    "ssh actions will not work.\n"
    "To use ssh, run 'pip install broker[{backend}]'."
)


def make_session(broker_settings=None, **kwargs):
    """Create a Session instance using the configured backend.

    Args:
        broker_settings: Optional settings object to use instead of global settings
        **kwargs: Additional arguments to pass to the Session constructor

    Returns:
        A Session instance from the configured backend
    """
    _settings = broker_settings or clone_global_settings()
    backend = _settings.SSH.BACKEND

    logger.debug(f"Attempting to load SSH backend: {backend}")

    try:
        # Look up the session class from entry points
        session_eps = entry_points(group="broker.ssh.session")
        # Get the specific backend requested or the first one registered
        session_cls = None
        for ep in session_eps:
            if ep.name == backend:
                session_cls = ep.load()
                break

        if not session_cls:
            backends = [ep.name for ep in session_eps]
            error_msg = f"SSH backend '{backend}' not supported. Supported backends: {backends}"
            logger.error(error_msg)
            raise ImportError(error_msg)

        # Create and return the session instance
        return session_cls(broker_settings=_settings, **kwargs)
    except ImportError:
        error_msg = SSH_NOT_INSTALLED_MSG.format(backend=backend)
        logger.warning(error_msg)

        # Create and return a dummy session that won't work but won't crash
        class DummySession:
            def __init__(self, broker_settings=None, **kwargs):
                self._settings = broker_settings or clone_global_settings()

        return DummySession(broker_settings=_settings, **kwargs)


def get_interactive_shell(broker_settings=None):
    """Get the InteractiveShell class for the configured backend.

    Args:
        broker_settings: Optional settings object to use instead of global settings

    Returns:
        The InteractiveShell class from the configured backend or None if not available
    """
    _settings = broker_settings or clone_global_settings()
    backend = _settings.SSH.BACKEND

    try:
        # Look up the interactive shell class from entry points
        shell_eps = entry_points(group="broker.ssh.interactive_shell")

        # Get the specific backend requested
        for ep in shell_eps:
            if ep.name == backend:
                return ep.load()

        # If no matching backend was found for interactive shell
        logger.debug(f"No interactive shell available for backend: {backend}")
        return None
    except ImportError:
        return None


# Expose the InteractiveShell for backwards compatibility
InteractiveShell = get_interactive_shell()

# Expose Session class (for backwards compatibility)
Session = make_session


class ContainerSession:
    """An approximation of ssh-based functionality from the Session class."""

    def __init__(self, cont_inst, runtime=None, broker_settings=None):
        self._cont_inst = cont_inst
        self._settings = broker_settings or clone_global_settings()
        if not runtime:
            runtime = self._settings.CONTAINER.runtime
        self.runtime = runtime

    def run(self, command, demux=True, **kwargs):
        """Container approximation of Session.run."""
        kwargs.pop("timeout", None)  # Timeouts are set at the client level
        kwargs["demux"] = demux
        if "'" in command:
            with tempfile.NamedTemporaryFile(mode="w", suffix=".sh") as tmp:
                tmp.write(command)
                tmp.seek(0)
                command = f"/bin/bash {tmp.name}"
                self.sftp_write(tmp.name)
        if any(s in command for s in "|&><"):
            # Containers don't handle pipes, redirects, etc well in a bare exec_run
            command = f"/bin/bash -c '{command}'"
        result = self._cont_inst._cont_inst.exec_run(command, **kwargs)
        if demux:
            result = helpers.Result.from_duplexed_exec(result, self.runtime)
        else:
            result = helpers.Result.from_nonduplexed_exec(result)
        return result

    def disconnect(self):
        """Needed for simple compatibility with Session."""

    @contextmanager
    def tail_file(self, filename):
        """Simulate tailing a file on the remote host."""
        initial_size = int(self.run(f"stat -c %s {filename}").stdout.strip())
        yield (res := helpers.Result())
        # get the contents of the file from the initial size to the end
        result = self.run(f"tail -c +{initial_size} {filename}")
        res.__dict__.update(result.__dict__)

    def sftp_write(self, source, destination=None, ensure_dir=True):
        """Add one of more files to the container."""
        # ensure source is a list of Path objects
        if not isinstance(source, list):
            source = [Path(source)]
        else:
            source = [Path(src) for src in source]
        # validate each source's existenence
        for src in source:
            if not Path(src).exists():
                raise FileNotFoundError(src)
        destination = destination or f"{source[0].parent}/"
        # Files need to be added to a tarfile
        with helpers.temporary_tar(source) as tar:
            logger.debug(f"{self._cont_inst.hostname} adding file(s) {source} to {destination}")
            if ensure_dir:
                if destination.endswith("/"):
                    self.run(f"mkdir -m 666 -p {destination}")
                else:
                    self.run(f"mkdir -m 666 -p {Path(destination).parent}")
            self._cont_inst._cont_inst.put_archive(str(destination), tar.read_bytes())

    def sftp_read(self, source, destination=None, return_data=False):
        """Get a file or directory from the container."""
        destination = Path(destination or source)
        logger.debug(f"{self._cont_inst.hostname} getting file {source}")
        data, status = self._cont_inst._cont_inst.get_archive(source)
        logger.debug(f"{self._cont_inst.hostname}: {status}")
        data = b"".join(d for d in data)
        if destination.name == "_raw":
            return data
        with helpers.data_to_tempfile(data, as_tar=True) as tar:
            del data
            if len(tar.getmembers()) == 1:
                f = tar.extractfile(tar.getmember(Path(source).name))
                if return_data:
                    logger.debug(f"Extracting {source}")
                    return f.read()
                else:
                    logger.debug(f"Extracting {source} to {destination}")
                    destination.parent.mkdir(parents=True, exist_ok=True)
                    destination.write_bytes(f.read())
            else:
                logger.warning("More than one member was found in the tar file.")
                tar.extractall(destination.parent if destination.is_file() else destination)

    def shell(self, pty=False):
        """Create and return an interactive shell instance."""
        raise NotImplementedError("ContainerSession.shell has not been implemented")



"""Broker settings module.

Useful items:
    settings: The settings object.
    INTERACTIVE_MODE: Whether or not Broker is running in interactive mode.
    BROKER_DIRECTORY: The directory where Broker looks for its files.
    settings_path: The path to the settings file.
    inventory_path: The path to the inventory file.
"""

import os
from pathlib import Path

import click
from dynaconf import Dynaconf, Validator
from dynaconf.validator import ValidationError

from broker.config_manager import ConfigManager
from broker.exceptions import ConfigurationError

INTERACTIVE_MODE = ConfigManager.interactive_mode
BROKER_DIRECTORY = Path.home().joinpath(".broker")

if "BROKER_DIRECTORY" in os.environ:
    envar_location = Path(os.environ["BROKER_DIRECTORY"])
    if envar_location.is_dir():
        BROKER_DIRECTORY = envar_location

# ensure the broker directory exists
BROKER_DIRECTORY.mkdir(parents=True, exist_ok=True)

settings_path = BROKER_DIRECTORY.joinpath("broker_settings.yaml")
inventory_path = BROKER_DIRECTORY.joinpath("inventory.yaml")


BASE_VALIDATORS = [
    Validator("SSH", is_type_of=dict, default={}),
    Validator("SSH.HOST_USERNAME", default="root"),
    Validator("SSH.HOST_PASSWORD", default="toor"),
    Validator("SSH.HOST_CONNECTION_TIMEOUT", default=60),
    Validator("SSH.HOST_SSH_PORT", default=22),
    Validator("SSH.HOST_SSH_KEY_FILENAME", default=None),
    Validator("SSH.HOST_IPV6", default=False),
    Validator("SSH.HOST_IPV4_FALLBACK", default=True),
    Validator("SSH.BACKEND", default="hussh"),
    Validator("LOGGING", is_type_of=dict),
    Validator(
        "LOGGING.CONSOLE_LEVEL",
        is_in=["error", "warning", "info", "debug", "trace", "silent"],
        default="info",
    ),
    Validator(
        "LOGGING.FILE_LEVEL",
        is_in=["error", "warning", "info", "debug", "trace", "silent"],
        default="debug",
    ),
    Validator("THREAD_LIMIT", default=None),
    Validator("INVENTORY_FIELDS", is_type_of=dict),
    Validator("INVENTORY_LIST_VARS", is_type_of=str, default="hostname | name"),
    Validator("LESS_COLORS", default=False),
]


def _handle_migrations(cfg_manager, file_exists):
    """Handle settings file migrations if needed."""
    if not (file_exists and cfg_manager._get_migrations()):
        return

    if INTERACTIVE_MODE:
        click.secho(
            "Broker settings file has pending migrations.\n"
            "Continuing without running the migrations may cause errors.",
            fg="red",
        )
        if click.confirm("Would you like to run the migrations now?"):
            cfg_manager.migrate()
        else:
            click.secho("Continuing without running migrations.", fg="yellow")
    else:
        cfg_manager.migrate()


def _handle_vault_env_vars():
    """Temporarily remove vault environment variables and return them."""
    vault_vars = {k: v for k, v in os.environ.items() if "VAULT_" in k}
    for k in vault_vars:
        del os.environ[k]
    return vault_vars


def _create_and_configure_settings(file_path, file_exists, config_dict):
    """Create settings object and apply configuration."""
    new_settings = Dynaconf(
        settings_file=str(file_path) if file_exists else None,
        ENVVAR_PREFIX_FOR_DYNACONF="BROKER",
        validators=BASE_VALIDATORS,
    )

    # Remove vault loader if set somehow
    new_settings._loaders = [loader for loader in new_settings._loaders if "vault" not in loader]

    # Add any configuration values passed in
    if config_dict:
        for key, value in config_dict.items():
            new_settings[key] = value

    return new_settings


def _validate_settings(new_settings, skip_validation, file_exists, file_path):
    """Validate settings and handle errors."""
    try:
        if skip_validation:
            new_settings.validators.validate(only="LOGGING")
        else:
            new_settings.validators.validate()
    except ValidationError as err:
        if file_exists:
            raise ConfigurationError(f"Configuration error in {file_path}: {err.args[0]}") from err
        # If no file exists, just ensure defaults are applied


def create_settings(
    config_dict=None, config_file=None, perform_migrations=False, skip_validation=False
):
    """Create a new settings object with custom configuration.

    Args:
        config_dict: Dictionary containing configuration values to overlay onto settings
        config_file: Path to a settings file to use instead of the default
        perform_migrations: Whether to check for and perform migrations
        skip_validation: Whether to skip all (but logging) validations

    Returns:
        A dynaconf settings object
    """
    file_path = config_file or settings_path
    file_exists = Path(file_path).exists() if file_path else False
    cfg_manager = ConfigManager(settings_path)

    # Handle migrations
    if perform_migrations:
        _handle_migrations(cfg_manager, file_exists)

    # Handle vault environment variables
    vault_vars = _handle_vault_env_vars()

    # Create and configure settings
    new_settings = _create_and_configure_settings(file_path, file_exists, config_dict)

    # Validate settings
    _validate_settings(new_settings, skip_validation, file_exists, file_path)

    # Restore environment variables
    os.environ.update(vault_vars)

    if not file_exists and not config_dict and not config_file and INTERACTIVE_MODE:
        click.secho(
            "No Broker settings file found. Using default settings.\n"
            "Run 'broker config init' to create a settings file.",
            fg="yellow",
        )

    return new_settings


def clone_global_settings():
    """Get broker settings by cloning the current global settings object."""
    return settings.dynaconf_clone()


class _SettingsProxy:
    """Proxy object that creates settings on first access."""

    def __init__(self):
        self._settings = None

    def _ensure_settings(self):
        if self._settings is None:
            self._settings = create_settings(perform_migrations=True)
        return self._settings

    def __getattr__(self, name):
        return getattr(self._ensure_settings(), name)

    def __getitem__(self, key):
        return self._ensure_settings()[key]

    def __setitem__(self, key, value):
        self._ensure_settings()[key] = value

    def __contains__(self, key):
        return key in self._ensure_settings()


# Create the global settings object (deferred)
settings = _SettingsProxy()



"""Module for Broker providers.

This module defines the `Provider` class, which is the base class for all Broker providers.
It provides useful methods for registering provider actions and must be inherited by all
Broker providers.

Attributes:
    PROVIDERS (dict): Dictionary of provider names and classes.
    PROVIDER_ACTIONS (dict): Dictionary of provider actions and their corresponding methods.
    PROVIDER_HELP (dict): Dictionary providing information to construct `broker providers --help`

Classes:
    Provider: Base class for all Broker providers.

Usage:
    To create a new Broker provider, create a new class that inherits from the `Provider` class
    and implements the required methods. For example:

    ```
    from broker.providers import Provider

    class MyProvider(Provider):
        def provider_help(self):
            # implementation here

        def get_inventory(self, **inventory_opts):
            # implementation here
    ```

Note: The `Provider` class should not be used directly.

"""

from abc import ABCMeta, abstractmethod
import inspect
from pathlib import Path

import dynaconf
from logzero import logger

from broker import exceptions
from broker.settings import clone_global_settings

# populate a list of all provider module names
_provider_imports = [
    f.stem for f in Path(__file__).parent.glob("*.py") if f.is_file() and f.stem != "__init__"
]

# ProviderName: ProviderClassObject
PROVIDERS = {}
# action: (InterfaceClass, "method_name")
PROVIDER_ACTIONS = {}
# action: (InterfaceClass, "method_name")
PROVIDER_HELP = []


class ProviderMeta(ABCMeta):
    """Metaclass that registers provider classes and actions."""

    def __new__(cls, name, bases, attrs):
        """Register provider classes and actions."""
        new_cls = super().__new__(cls, name, bases, attrs)
        if name != "Provider":
            PROVIDERS[name] = new_cls
            logger.debug(f"Registered provider {name}")
            for attr, obj in attrs.items():
                if attr == "provider_help":
                    # register the help options based on the function arguments
                    for _name, param in inspect.signature(obj).parameters.items():
                        if _name not in ("self", "kwargs", "broker_settings"):
                            # (name, prov_cls, is_flag, help_override)
                            PROVIDER_HELP.append(
                                (
                                    _name,
                                    new_cls,
                                    isinstance(param.default, bool),
                                    getattr(obj, "_help_overrides", {}).get(_name),
                                )
                            )
                            logger.debug(f"Registered help option {_name} for provider {name}")
                elif hasattr(obj, "_as_action"):
                    for action in obj._as_action:
                        PROVIDER_ACTIONS[action] = (new_cls, attr)
                        logger.debug(f"Registered action {action} for provider {name}")
        return new_cls


class Provider(metaclass=ProviderMeta):
    """Abstract base class for all providers.

    This class should be subclassed by all provider implementations. It provides a
    metaclass that registers provider classes and actions.

    Attributes:
        _validators (list): A list of Dynaconf Validators specific to the provider.
        _checkout_options (list): A list of checkout options to add to each command.
        _execute_options (list): A list of execute options to add to each command.
        _fresh_settings (dynaconf.Dynaconf): A clone of the global settings object.
        _sensitive_attrs (list): A list of sensitive attributes that should not be logged.
    """

    # Populate with a list of Dynaconf Validators specific to your provider
    _validators = []
    # Populate these to add your checkout and execute options to each command
    # _checkout_options = [click.option("--workflow", type=str, help="Help text")]
    _checkout_options = []
    _execute_options = []
    _extend_options = []
    _sensitive_attrs = []

    def __init__(self, broker_settings=None, **kwargs):
        cls_name = self.__class__.__name__
        if broker_settings:
            self._settings = broker_settings
            # Add provider validators to custom settings
            if validators := getattr(self.__class__, "_validators", None):
                self._settings.validators.extend(validators)
            logger.debug(
                f"{cls_name} provider using custom settings object: {broker_settings.to_dict()}"
            )
        else:
            self._settings = clone_global_settings()
            logger.debug(f"{cls_name} provider using global settings")
        self._construct_params = []
        logger.debug(f"{cls_name} provider instantiated with {kwargs=}")
        self.instance = kwargs.pop(f"{cls_name}", None)
        self._validate_settings(self.instance)

    def _validate_settings(self, instance_name=None):
        """Load and validate provider settings.

        Each provider's settings can include an instances dict with specific instance
        details.
        One instance should have a "default" key set to True, if instances are defined.
        General provider settings should live on the top level for that provider.

        :param instance_name: A string matching an instance name
        """
        section_name = self.__class__.__name__.upper()
        # if the provider has instances, load the instance settings
        if self._settings.get(section_name, {}).get("instances"):
            fresh_settings = self._settings.get(section_name).copy()
            instance_name = instance_name or getattr(self, "instance", None)
            # first check to see if we have a direct match
            if not (instance_values := fresh_settings.instances.get(instance_name)):
                # if no direct match is found, or no instance is provided, find the default
                for name, values in fresh_settings.instances.items():
                    if values.get("default") or len(fresh_settings.instances) == 1:
                        instance_name, instance_values = name, values
                        break
            self.instance = instance_name  # store the instance name on the provider
            fresh_settings.update(instance_values)
            self._settings[section_name] = fresh_settings
            if not instance_values.get("override_envars"):
                # if a provider instance doesn't want to override envars, load them
                self._settings.execute_loaders(loaders=[dynaconf.loaders.env_loader])
        # add our validators
        self._settings.validators.extend(self._validators)
        # use selective validation to only validate the instance settings
        try:
            self._settings.validators.validate(only=section_name)
        except dynaconf.ValidationError as err:
            raise exceptions.ConfigurationError(err) from err

    def _set_attributes(self, obj, attrs):
        obj.__dict__.update(attrs)

    @abstractmethod
    def construct_host(self, host_cls, provider_params, **kwargs):
        """Construct a host object from a host class and include relevent provider params."""

    @abstractmethod
    def provider_help(self, broker_settings=None):
        """Help options that will be added to the CLI.

        Anything other than 'self', 'kwargs', and 'broker_settings' will be added as a help option
        To specify a flag, set the default value to a boolean
        Everything else should default to None
        """

    @abstractmethod
    def get_inventory(self, **kwargs):
        """Pull inventory information from the provider."""

    @abstractmethod
    def extend(self):
        """Extend the reservation of a host. Not all providers support this."""

    @abstractmethod
    def release(self, host_obj):
        """Release/return a host to the provider. Often this is a deletion or removal."""

    def __repr__(self):
        """Return a string representation of the provider."""
        inner = ", ".join(
            f"{k}={'******' if k in self._sensitive_attrs and v else v}"
            for k, v in self.__dict__.items()
            if not k.startswith("_") and not callable(v)
        )
        return f"{self.__class__.__name__}({inner})"

    @staticmethod
    def register_action(*as_names):
        """Decorate a provider method to register it as a provider action.

        :param as_names: One or more action names to register the decorated function as
        """

        def decorator(func):
            func._as_action = as_names or [func.__name__]
            return func

        return decorator

    @staticmethod
    def help_override(**overrides):
        """Decorate the provider_help function to overwrite auto-generated help text."""

        def wrapper(func):
            func._help_overrides = overrides
            return func

        return wrapper



"""Config migrations for versions older than 0.6.0 to 0.6.0."""

from logzero import logger

TO_VERSION = "0.6.0"


def migrate_instances(config_dict):
    """Migrate instances from a list of dicts to a dict of dicts."""
    logger.debug("Migrating instances from a list to a dict.")
    for key, val in config_dict.items():
        if not isinstance(val, dict):
            continue
        if "instances" in val and isinstance(val["instances"], list):
            old_instances = val.pop("instances")
            val["instances"] = {}
            for inst in old_instances:
                val["instances"].update(inst)
        config_dict[key] = val
    return config_dict


def remove_testprovider(config_dict):
    """Remove the testprovider from the config."""
    logger.debug("Removing the testprovider from the config.")
    config_dict.pop("TestProvider", None)
    return config_dict


def remove_test_nick(config_dict):
    """Remove the test nick from the config."""
    logger.debug("Removing the test nick from the config.")
    nicks = config_dict.get("nicks", {})
    nicks.pop("test_nick", None)
    config_dict["nicks"] = nicks
    return config_dict


def move_ssh_settings(config_dict):
    """Move SSH settings from the top level into its own chunk."""
    # Check if the migration has already been performed
    if "ssh" in config_dict:
        return config_dict
    logger.debug("Moving SSH settings into their own section.")
    ssh_settings = {
        "backend": config_dict.pop("ssh_backend", "ssh2-python312"),
        "host_username": config_dict.pop("host_username", "root"),
        "host_password": config_dict.pop("host_password", "toor"),
        "host_ipv6": config_dict.pop("host_ipv6", False),
        "host_ipv4_fallback": config_dict.pop("host_ipv4_fallback", True),
    }
    if ssh_port := config_dict.pop("host_ssh_port", None):
        ssh_settings["ssh_port"] = ssh_port
    if ssh_key := config_dict.pop("host_ssh_key_filename", None):
        ssh_settings["host_ssh_key_filename"] = ssh_key
    config_dict["ssh"] = ssh_settings
    return config_dict


def add_thread_limit(config_dict):
    """Add a thread limit to the config."""
    logger.debug("Adding a thread limit to the config.")
    config_dict["thread_limit"] = config_dict.get("thread_limit")
    return config_dict


def add_inventory_fields(config_dict):
    """Inventory fields are new in this version.

    Example:
        # Customize the fields and values presented by `broker inventory`
        # Almost all field values should correspond to a field in your Broker inventory
        inventory_fields:
        Host: hostname | name  # use a | to allow fallback values
        Provider: _broker_provider  # just pull the _broker_provider value
        Action: $action  # some special field values are possible, check the wiki
        OS: os_distribution os_distribution_version  # you can combine multiple values with a space between
    """
    if "inventory_fields" in config_dict:
        return config_dict
    logger.debug("Adding inventory fields to the config.")
    config_dict["inventory_fields"] = {
        "Host": "hostname",
        "Provider": "_broker_provider",
        "Action": "$action",
        "OS": "os_distribution os_distribution_version",
    }
    config_dict["inventory_list_vars"] = "hostname | name"
    return config_dict


def add_color_control(config_dict):
    """Add in the new `less_colors` field."""
    logger.debug("Adding the less_colors field to the config.")
    config_dict["less_colors"] = config_dict.get("less_colors", False)
    return config_dict


def run_migrations(config_dict):
    """Run all migrations."""
    logger.info(f"Running config migrations for {TO_VERSION}.")
    config_dict = migrate_instances(config_dict)
    config_dict = remove_testprovider(config_dict)
    config_dict = remove_test_nick(config_dict)
    config_dict = move_ssh_settings(config_dict)
    config_dict = add_thread_limit(config_dict)
    config_dict = add_inventory_fields(config_dict)
    config_dict = add_color_control(config_dict)
    config_dict["_version"] = TO_VERSION
    return config_dict
